"""Definition of non-processing subcommands."""

from itertools import zip_longest
from math import ceil
from shutil import get_terminal_size
from textwrap import indent, TextWrapper
import sys

import loam.tools
import pandas

from . import conf, phyvars, __version__
from . import stagyydata
from .config import CONFIG_FILE, CONFIG_LOCAL
from .misc import baredoc


def info_cmd():
    """Print basic information about StagYY run.

    Other Parameters:
        conf.info
    """
    varlist = [var for var in conf.info.output.replace(',', ' ').split()]
    sdat = stagyydata.StagyyData()
    lsnap = sdat.snaps[-1]
    lstep = sdat.steps[-1]
    print(f'StagYY run in {sdat.path}')
    if lsnap.geom.threed:
        dimension = '{0.nxtot} x {0.nytot} x {0.nztot}'.format(lsnap.geom)
    elif lsnap.geom.twod_xz:
        dimension = '{0.nxtot} x {0.nztot}'.format(lsnap.geom)
    else:
        dimension = '{0.nytot} x {0.nztot}'.format(lsnap.geom)
    if lsnap.geom.cartesian:
        print('Cartesian', dimension)
    elif lsnap.geom.cylindrical:
        print('Cylindrical', dimension)
    else:
        print('Spherical', dimension)
    print()
    for step in sdat.walk:
        print(f'Step {step.istep}/{lstep.istep}', end='')
        if step.isnap is not None:
            print(f', snapshot {step.isnap}/{lsnap.isnap}')
        else:
            print()
        series = step.timeinfo.loc[varlist]
        if conf.scaling.dimensional:
            series = series.copy()
            dimensions = []
            for var, val in series.iteritems():
                dim = phyvars.TIME.get(var) or phyvars.TIME_EXTRA.get(var)
                dim = dim.dim if dim is not None else '1'
                if dim == '1':
                    dimensions.append('')
                else:
                    series[var], dim = sdat.scale(val, dim)
                    dimensions.append(dim)
            series = pandas.concat(
                [series, pandas.Series(data=dimensions, index=series.index,
                                       name='dim')],
                axis=1)
        print(indent(series.to_string(header=False), '  '))
        print()


def _pretty_print(key_val, sep=': ', min_col_width=39, text_width=None):
    """Print a iterable of key/values.

    Args:
        key_val (list of (str, str)): the pairs of section names and text.
        sep (str): separator between section names and text.
        min_col_width (int): minimal acceptable column width
        text_width (int): text width to use. If set to None, will try to infer
            the size of the terminal.
    """
    if text_width is None:
        text_width = get_terminal_size().columns
    if text_width < min_col_width:
        min_col_width = text_width
    ncols = (text_width + 1) // (min_col_width + 1)
    colw = (text_width + 1) // ncols - 1
    ncols = min(ncols, len(key_val))

    wrapper = TextWrapper(width=colw)
    lines = []
    for key, val in key_val:
        if len(key) + len(sep) >= colw // 2:
            wrapper.subsequent_indent = ' '
        else:
            wrapper.subsequent_indent = ' ' * (len(key) + len(sep))
        lines.extend(wrapper.wrap(f'{key}{sep}{val}'))

    chunks = []
    for rem_col in range(ncols, 1, -1):
        isep = ceil(len(lines) / rem_col)
        while isep < len(lines) and lines[isep][0] == ' ':
            isep += 1
        chunks.append(lines[:isep])
        lines = lines[isep:]
    chunks.append(lines)
    lines = zip_longest(*chunks, fillvalue='')

    fmt = '|'.join([f'{{:{colw}}}'] * (ncols - 1))
    fmt += '|{}' if ncols > 1 else '{}'
    print(*(fmt.format(*line) for line in lines), sep='\n')


def _layout(dict_vars, dict_vars_extra):
    """Print nicely [(var, description)] from phyvars."""
    desc = [(v, m.description) for v, m in dict_vars.items()]
    desc.extend((v, baredoc(m.description))
                for v, m in dict_vars_extra.items())
    _pretty_print(desc, min_col_width=26)


def var_cmd():
    """Print a list of available variables.

    See :mod:`stagpy.phyvars` where the lists of variables organized by command
    are defined.
    """
    print_all = not any(val for _, val in conf.var.opt_vals_())
    if print_all or conf.var.field:
        print('field:')
        _layout(phyvars.FIELD, phyvars.FIELD_EXTRA)
        print()
    if print_all or conf.var.sfield:
        print('surface field:')
        _layout(phyvars.SFIELD, {})
        print()
    if print_all or conf.var.rprof:
        print('rprof:')
        _layout(phyvars.RPROF, phyvars.RPROF_EXTRA)
        print()
    if print_all or conf.var.time:
        print('time:')
        _layout(phyvars.TIME, phyvars.TIME_EXTRA)
        print()
    if print_all or conf.var.refstate:
        print('refstate:')
        _layout(phyvars.REFSTATE, {})
        print()
    if print_all or conf.var.plates:
        print('plates:')
        _layout(phyvars.PLATES, {})


def version_cmd():
    """Print StagPy version.

    Use :data:`stagpy.__version__` to obtain the version in a script.
    """
    print(f'stagpy version: {__version__}')


def report_parsing_problems(parsing_out):
    """Output message about potential parsing problems."""
    _, empty, faulty = parsing_out
    if CONFIG_FILE in empty or CONFIG_FILE in faulty:
        print('Unable to read global config file', CONFIG_FILE,
              file=sys.stderr)
        print('Please run stagpy config --create',
              sep='\n', end='\n\n', file=sys.stderr)
    if CONFIG_LOCAL in faulty:
        print('Unable to read local config file', CONFIG_LOCAL,
              file=sys.stderr)
        print('Please run stagpy config --create_local',
              sep='\n', end='\n\n', file=sys.stderr)


def config_pp(subs):
    """Pretty print of configuration options.

    Args:
        subs (iterable of str): iterable with the list of conf sections to
            print.
    """
    print('(c|f): available only as CLI argument/in the config file',
          end='\n\n')
    for sub in subs:
        hlp_lst = []
        for opt, meta in conf[sub].defaults_():
            if meta.cmd_arg ^ meta.conf_arg:
                opt += ' (c)' if meta.cmd_arg else ' (f)'
            hlp_lst.append((opt, meta.help))
        if hlp_lst:
            print(f'{sub}:')
            _pretty_print(hlp_lst, sep=' -- ',
                          text_width=min(get_terminal_size().columns, 100))
            print()


def config_cmd():
    """Configuration handling.

    Other Parameters:
        conf.config
    """
    if not (conf.common.config or conf.config.create or
            conf.config.create_local or conf.config.update or
            conf.config.edit):
        config_pp(conf.sections_())
    loam.tools.config_cmd_handler(conf)



import re
import pytest
import stagpy.args


def test_no_args(capsys):
    stagpy.args.parse_args([])()
    output = capsys.readouterr()
    expected = re.compile(
        r'StagPy is a tool to.*'
        r'Run `stagpy -h` for usage\n$',
        flags=re.DOTALL)
    assert expected.fullmatch(output.out)


def test_help(capsys):
    with pytest.raises(SystemExit):
        stagpy.args.parse_args(['-h'])
    output = capsys.readouterr()
    expected = re.compile(
        r'^usage:.*\nStagPy is a tool to.*'
        r'\npositional arguments.*\noptional arguments.*$',
        flags=re.DOTALL)
    assert expected.fullmatch(output.out)


def test_invalid_argument(capsys):
    with pytest.raises(SystemExit):
        stagpy.args.parse_args(['-dummyinvalidarg'])
    output = capsys.readouterr()
    expected = re.compile(
        r'^usage: .*error: unrecognized arguments:.*\n$',
        flags=re.DOTALL)
    assert expected.fullmatch(output.err)


def test_invalid_subcmd(capsys):
    with pytest.raises(SystemExit):
        stagpy.args.parse_args(['dummyinvalidcmd'])
    output = capsys.readouterr()
    expected = re.compile(
        r'^usage: .*error:.*invalid choice:.*\n$',
        flags=re.DOTALL)
    assert expected.fullmatch(output.err)


def test_field_subcmd():
    func = stagpy.args.parse_args(['field'])
    assert func is stagpy.field.cmd


def test_rprof_subcmd():
    func = stagpy.args.parse_args(['rprof'])
    assert func is stagpy.rprof.cmd


def test_time_cmd():
    func = stagpy.args.parse_args(['time'])
    assert func is stagpy.time_series.cmd


def test_plates_subcmd():
    func = stagpy.args.parse_args(['plates'])
    assert func is stagpy.plates.cmd


def test_info_subcmd():
    func = stagpy.args.parse_args(['info'])
    assert func is stagpy.commands.info_cmd


def test_var_subcmd():
    func = stagpy.args.parse_args(['var'])
    assert func is stagpy.commands.var_cmd


def test_version_subcmd():
    func = stagpy.args.parse_args(['version'])
    assert func is stagpy.commands.version_cmd


def test_config_subcmd():
    func = stagpy.args.parse_args(['config'])
    assert func is stagpy.commands.config_cmd



"""The stagpy module is callable."""

import importlib
import signal
import sys
import warnings

from . import DEBUG, sigint_handler


def main():
    """Implement StagPy entry point."""
    if not DEBUG:
        signal.signal(signal.SIGINT, sigint_handler)
        warnings.simplefilter('ignore')
    args = importlib.import_module('stagpy.args')
    error = importlib.import_module('stagpy.error')
    try:
        args.parse_args()()
    except error.StagpyError as err:
        if DEBUG:
            raise
        errtype = type(err).__name__
        print('Oops! StagPy encountered the following problem while '
              'processing your request.', 'Please check the path to your '
              'simulation and the command line arguments.', '',
              f'{errtype}: {err}', sep='\n', file=sys.stderr)
        sys.exit()


if __name__ == '__main__':
    main()



"""Implementation of Step objects.

Note:
    This module and the classes it defines are internals of StagPy, they
    should not be used in an external script.  Instead, use the
    :class:`~stagpy.stagyydata.StagyyData` class.
"""

from collections.abc import Mapping
from collections import namedtuple
from itertools import chain

import numpy as np

from . import error, misc, phyvars, stagyyparsers
from .misc import CachedReadOnlyProperty as crop


class _Geometry:
    """Geometry information.

    It is deduced from the information in the header of binary field files
    output by StagYY.
    """

    def __init__(self, header, step):
        self._header = header
        self._step = step
        self._shape = {'sph': False, 'cyl': False, 'axi': False,
                       'ntot': list(header['nts']) + [header['ntb']]}
        self._init_shape()

    def _scale_radius_mo(self, radius):
        """Rescale radius for MO runs."""
        if self._step.sdat.par['magma_oceans_in']['magma_oceans_mode']:
            return self._header['mo_thick_sol'] * (
                radius + self._header['mo_lambda'])
        return radius

    @crop
    def nttot(self):
        """Number of grid point along the x/theta direction."""
        return self._shape['ntot'][0]

    @crop
    def nptot(self):
        """Number of grid point along the y/phi direction."""
        return self._shape['ntot'][1]

    @crop
    def nrtot(self):
        """Number of grid point along the z/r direction."""
        return self._shape['ntot'][2]

    @crop
    def nbtot(self):
        """Number of blocks."""
        return self._shape['ntot'][3]

    nxtot = nttot
    nytot = nptot
    nztot = nrtot

    @crop
    def r_walls(self):
        """Position of FV walls along the z/r direction."""
        rgeom = self._header.get("rgeom")
        if rgeom is not None:
            walls = rgeom[:, 0] + self.rcmb
        else:
            walls = self._header["e3_coord"] + self.rcmb
            walls.append(self._step.rprofs.bounds[1])
        return self._scale_radius_mo(walls)

    @crop
    def r_centers(self):
        """Position of FV centers along the z/r direction."""
        rgeom = self._header.get("rgeom")
        if rgeom is not None:
            walls = rgeom[:-1, 1] + self.rcmb
        else:
            walls = self._step.rprofs.centers
        return self._scale_radius_mo(walls)

    @crop
    def t_walls(self):
        """Position of FV walls along x/theta."""
        if self.threed or self.twod_xz:
            if self.yinyang:
                tmin, tmax = -np.pi / 4, np.pi / 4
            elif self.curvilinear:
                # should take theta_position/theta_center into account
                tmin = 0
                tmax = min(np.pi,
                           self._step.sdat.par['geometry']['aspect_ratio'][0])
            else:
                tmin = 0
                tmax = self._step.sdat.par['geometry']['aspect_ratio'][0]
            return np.linspace(tmin, tmax, self.nttot + 1)
        # twoD YZ
        center = np.pi / 2 if self.curvilinear else 0
        d_t = (self.p_walls[1] - self.p_walls[0]) / 2
        return np.array([center - d_t, center + d_t])

    @crop
    def t_centers(self):
        """Position of FV centers along x/theta."""
        return (self.t_walls[:-1] + self.t_walls[1:]) / 2

    @crop
    def p_walls(self):
        """Position of FV walls along y/phi."""
        if self.threed or self.twod_yz:
            if self.yinyang:
                pmin, pmax = -3 * np.pi / 4, 3 * np.pi / 4
            elif self.curvilinear:
                pmin = 0
                pmax = min(2 * np.pi,
                           self._step.sdat.par['geometry']['aspect_ratio'][1])
            else:
                pmin = 0
                pmax = self._step.sdat.par['geometry']['aspect_ratio'][1]
            return np.linspace(pmin, pmax, self.nptot + 1)
        # twoD YZ
        d_p = (self.t_walls[1] - self.t_walls[0]) / 2
        return np.array([-d_p, d_p])

    @crop
    def p_centers(self):
        """Position of FV centers along y/phi."""
        return (self.p_walls[:-1] + self.p_walls[1:]) / 2

    z_walls = r_walls
    z_centers = r_centers
    x_walls = t_walls
    x_centers = t_centers
    y_walls = p_walls
    y_centers = p_centers

    def _init_shape(self):
        """Determine shape of geometry."""
        shape = self._step.sdat.par['geometry']['shape'].lower()
        aspect = self._header['aspect']
        if self._header['rcmb'] is not None and self._header['rcmb'] >= 0:
            # curvilinear
            self._shape['cyl'] = self.twod_xz and (shape == 'cylindrical' or
                                                   aspect[0] >= np.pi)
            self._shape['sph'] = not self._shape['cyl']
        elif self._header['rcmb'] is None:
            self._header['rcmb'] = self._step.sdat.par['geometry']['r_cmb']
            if self._header['rcmb'] >= 0:
                if self.twod_xz and shape == 'cylindrical':
                    self._shape['cyl'] = True
                elif shape == 'spherical':
                    self._shape['sph'] = True
        self._shape['axi'] = self.cartesian and self.twod_xz and \
            shape == 'axisymmetric'

    @crop
    def rcmb(self):
        """Radius of CMB, 0 in cartesian geometry."""
        return max(self._header["rcmb"], 0)

    @property
    def cartesian(self):
        """Whether the grid is in cartesian geometry."""
        return not self.curvilinear

    @property
    def curvilinear(self):
        """Whether the grid is in curvilinear geometry."""
        return self.spherical or self.cylindrical

    @property
    def cylindrical(self):
        """Whether the grid is in cylindrical geometry (2D spherical)."""
        return self._shape['cyl']

    @property
    def spherical(self):
        """Whether the grid is in spherical geometry."""
        return self._shape['sph']

    @property
    def yinyang(self):
        """Whether the grid is in Yin-yang geometry (3D spherical)."""
        return self.spherical and self.nbtot == 2

    @property
    def twod_xz(self):
        """Whether the grid is in the XZ plane only."""
        return self.nytot == 1

    @property
    def twod_yz(self):
        """Whether the grid is in the YZ plane only."""
        return self.nxtot == 1

    @property
    def twod(self):
        """Whether the grid is 2 dimensional."""
        return self.twod_xz or self.twod_yz

    @property
    def threed(self):
        """Whether the grid is 3 dimensional."""
        return not self.twod

    def at_z(self, zval):
        """Return iz closest to given zval position.

        In spherical geometry, the bottom boundary is considered to be at z=0.
        Use :meth:`at_r` to find a cell at a given radial position.
        """
        if self.curvilinear:
            zval += self.rcmb
        return np.argmin(np.abs(self.z_centers - zval))

    def at_r(self, rval):
        """Return ir closest to given rval position.

        If called in cartesian geometry, this is equivalent to :meth:`at_z`.
        """
        return np.argmin(np.abs(self.r_centers - rval))


class _Fields(Mapping):
    """Fields data structure.

    The :attr:`Step.fields` attribute is an instance of this class.

    :class:`_Fields` inherits from :class:`collections.abc.Mapping`. Keys are
    fields names defined in :data:`stagpy.phyvars.[S]FIELD[_EXTRA]`.

    Attributes:
        step (:class:`Step`): the step object owning the :class:`_Fields`
            instance.
    """

    def __init__(self, step, variables, extravars, files, filesh5):
        self.step = step
        self._vars = variables
        self._extra = extravars
        self._files = files
        self._filesh5 = filesh5
        self._data = {}
        super().__init__()

    def __getitem__(self, name):
        if name in self._data:
            return self._data[name]
        if name in self._vars:
            fld_names, parsed_data = self._get_raw_data(name)
        elif name in self._extra:
            self._data[name] = self._extra[name].description(self.step)
            return self._data[name]
        else:
            raise error.UnknownFieldVarError(name)
        if parsed_data is None:
            raise error.MissingDataError(
                f'Missing field {name} in step {self.step.istep}')
        header, fields = parsed_data
        self._cropped__header = header
        for fld_name, fld in zip(fld_names, fields):
            self._set(fld_name, fld)
        return self._data[name]

    def __iter__(self):
        return (fld for fld in chain(self._vars, self._extra)
                if fld in self)

    def __contains__(self, item):
        try:
            return self[item] is not None
        except error.StagpyError:
            return False

    def __len__(self):
        return len(iter(self))

    def __eq__(self, other):
        return self is other

    def _get_raw_data(self, name):
        """Find file holding data and return its content."""
        # try legacy first, then hdf5
        filestem = ''
        for filestem, list_fvar in self._files.items():
            if name in list_fvar:
                break
        fieldfile = self.step.sdat.filename(filestem, self.step.isnap,
                                            force_legacy=True)
        if not fieldfile.is_file():
            fieldfile = self.step.sdat.filename(filestem, self.step.isnap)
        parsed_data = None
        if fieldfile.is_file():
            parsed_data = stagyyparsers.fields(fieldfile)
        elif self.step.sdat.hdf5 and self._filesh5:
            # files in which the requested data can be found
            files = [(stem, fvars) for stem, fvars in self._filesh5.items()
                     if name in fvars]
            for filestem, list_fvar in files:
                if filestem in phyvars.SFIELD_FILES_H5:
                    xmff = 'Data{}.xmf'.format(
                        'Bottom' if name.endswith('bot') else 'Surface')
                    header = self._header
                else:
                    xmff = 'Data.xmf'
                    header = None
                parsed_data = stagyyparsers.read_field_h5(
                    self.step.sdat.hdf5 / xmff, filestem,
                    self.step.isnap, header)
                if parsed_data is not None:
                    break
        return list_fvar, parsed_data

    def _set(self, name, fld):
        sdat = self.step.sdat
        col_fld = sdat._collected_fields
        col_fld.append((self.step.istep, name))
        if sdat.nfields_max is not None:
            while len(col_fld) > sdat.nfields_max:
                istep, fld_name = col_fld.pop(0)
                del sdat.steps[istep].fields[fld_name]
        self._data[name] = fld

    def __delitem__(self, name):
        if name in self._data:
            del self._data[name]

    @crop
    def _header(self):
        binfiles = self.step.sdat._binfiles_set(self.step.isnap)
        if binfiles:
            return stagyyparsers.fields(binfiles.pop(), only_header=True)
        elif self.step.sdat.hdf5:
            xmf = self.step.sdat.hdf5 / 'Data.xmf'
            return stagyyparsers.read_geom_h5(xmf, self.step.isnap)[0]

    @crop
    def geom(self):
        """Geometry information.

        :class:`_Geometry` instance holding geometry information. It is
        issued from binary files holding field information. It is set to
        None if not available for this time step.
        """
        if self._header is not None:
            return _Geometry(self._header, self.step)


class _Tracers:
    """Tracers data structure.

    The :attr:`Step.tracers` attribute is an instance of this class.

    :class:`_Tracers` implements the getitem mechanism. Items are tracervar
    names such as 'Type' or 'Mass'.  The position of tracers are the 'x', 'y'
    and 'z' items.

    Attributes:
        step (:class:`Step`): the step object owning the :class:`_Tracers`
            instance.
    """

    def __init__(self, step):
        self.step = step
        self._data = {}

    def __getitem__(self, name):
        if name in self._data:
            return self._data[name]
        data = stagyyparsers.tracers(
            self.step.sdat.filename('tra', timestep=self.step.isnap,
                                    force_legacy=True))
        if data is None and self.step.sdat.hdf5:
            position = any(axis not in self._data for axis in 'xyz')
            self._data.update(
                stagyyparsers.read_tracers_h5(
                    self.step.sdat.hdf5 / 'DataTracers.xmf', name,
                    self.step.isnap, position))
        elif data is not None:
            self._data.update(data)
        if name not in self._data:
            self._data[name] = None
        return self._data[name]

    def __iter__(self):
        raise TypeError('tracers collection is not iterable')


Rprof = namedtuple('Rprof', ['values', 'rad', 'meta'])


class _Rprofs:
    """Radial profiles data structure.

    The :attr:`Step.rprofs` attribute is an instance of this class.

    :class:`_Rprofs` implements the getitem mechanism.  Keys are profile names
    defined in :data:`stagpy.phyvars.RPROF[_EXTRA]`.  An item is a named tuple
    ('values', 'rad', 'meta'), respectively the profile itself, the radial
    position at which it is evaluated, and meta is a
    :class:`stagpy.phyvars.Varr` instance with relevant metadata.  Note that
    profiles are automatically scaled if conf.scaling.dimensional is True.

    Attributes:
        step (:class:`Step`): the step object owning the :class:`_Rprofs`
            instance
    """

    def __init__(self, step):
        self.step = step
        self._cached_extra = {}

    @crop
    def _data(self):
        step = self.step
        return step.sdat._rprof_and_times[0].get(step.istep)

    @property
    def _rprofs(self):
        if self._data is None:
            step = self.step
            raise error.MissingDataError(
                f'No rprof data in step {step.istep} of {step.sdat}')
        return self._data

    def __getitem__(self, name):
        step = self.step
        if name in self._rprofs.columns:
            rprof = self._rprofs[name].values
            rad = self.centers
            if name in phyvars.RPROF:
                meta = phyvars.RPROF[name]
            else:
                meta = phyvars.Varr(name, None, '1')
        elif name in self._cached_extra:
            rprof, rad, meta = self._cached_extra[name]
        elif name in phyvars.RPROF_EXTRA:
            meta = phyvars.RPROF_EXTRA[name]
            rprof, rad = meta.description(step)
            meta = phyvars.Varr(misc.baredoc(meta.description),
                                meta.kind, meta.dim)
            self._cached_extra[name] = rprof, rad, meta
        else:
            raise error.UnknownRprofVarError(name)
        rprof, _ = step.sdat.scale(rprof, meta.dim)
        rad, _ = step.sdat.scale(rad, 'm')

        return Rprof(rprof, rad, meta)

    @crop
    def centers(self):
        """Radial position of cell centers."""
        return self._rprofs['r'].values + self.bounds[0]

    @crop
    def walls(self):
        """Radial position of cell walls."""
        rbot, rtop = self.bounds
        try:
            walls = self.step.fields.geom.r_walls
        except error.StagpyError:
            # assume walls are mid-way between T-nodes
            # could be T-nodes at center between walls
            centers = self.centers
            walls = (centers[:-1] + centers[1:]) / 2
            walls = np.insert(walls, 0, rbot)
            walls = np.append(walls, rtop)
        return walls

    @crop
    def bounds(self):
        """Radial or vertical position of boundaries.

        Radial/vertical positions of boundaries of the domain.
        """
        step = self.step
        if step.geom is not None:
            rcmb = step.geom.rcmb
        else:
            rcmb = step.sdat.par['geometry']['r_cmb']
            if step.sdat.par['geometry']['shape'].lower() == 'cartesian':
                rcmb = 0
        rbot = max(rcmb, 0)
        thickness = (step.sdat.scales.length
                     if step.sdat.par['switches']['dimensional_units'] else 1)
        return rbot, rbot + thickness


class Step:
    """Time step data structure.

    Elements of :class:`_Steps` and :class:`_Snaps` instances are all
    :class:`Step` instances. Note that :class:`Step` objects are not
    duplicated.

    Examples:
        Here are a few examples illustrating some properties of :class:`Step`
        instances.

        >>> sdat = StagyyData('path/to/run')
        >>> istep_last_snap = sdat.snaps[-1].istep
        >>> assert(sdat.steps[istep_last_snap] is sdat.snaps[-1])
        >>> n = 0  # or any valid time step / snapshot index
        >>> assert(sdat.steps[n].sdat is sdat)
        >>> assert(sdat.steps[n].istep == n)
        >>> assert(sdat.snaps[n].isnap == n)
        >>> assert(sdat.steps[n].geom is sdat.steps[n].fields.geom)
        >>> assert(sdat.snaps[n] is sdat.snaps[n].fields.step)

    Args:
        istep (int): the index of the time step that the instance
            represents.
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance owning the :class:`Step` instance.

    Attributes:
        istep (int): the index of the time step that the instance
            represents.
        sdat (:class:`~stagpy.stagyydata.StagyyData`): the StagyyData
            instance owning the :class:`Step` instance.
        fields (:class:`_Fields`): fields available at this time step.
        sfields (:class:`_Fields`): surface fields available at this time
            step.
        tracers (:class:`_Tracers`): tracers available at this time step.
    """

    def __init__(self, istep, sdat):
        self.istep = istep
        self.sdat = sdat
        self.fields = _Fields(self, phyvars.FIELD, phyvars.FIELD_EXTRA,
                              phyvars.FIELD_FILES, phyvars.FIELD_FILES_H5)
        self.sfields = _Fields(self, phyvars.SFIELD, [],
                               phyvars.SFIELD_FILES, phyvars.SFIELD_FILES_H5)
        self.tracers = _Tracers(self)
        self.rprofs = _Rprofs(self)
        self._isnap = -1

    def __repr__(self):
        if self.isnap is not None:
            return f'{self.sdat!r}.snaps[{self.isnap}]'
        else:
            return f'{self.sdat!r}.steps[{self.istep}]'

    @property
    def geom(self):
        """Geometry information.

        :class:`_Geometry` instance holding geometry information. It is
        issued from binary files holding field information. It is set to
        None if not available for this time step.
        """
        return self.fields.geom

    @property
    def timeinfo(self):
        """Time series data of the time step."""
        try:
            info = self.sdat.tseries.at_step(self.istep)
        except KeyError:
            raise error.MissingDataError(f'No time series for {self!r}')
        return info

    @property
    def time(self):
        """Time of this time step."""
        steptime = None
        try:
            steptime = self.timeinfo['t']
        except error.MissingDataError:
            if self.isnap is not None:
                steptime = self.geom.ti_ad
        return steptime

    @property
    def isnap(self):
        """Snapshot index corresponding to time step.

        It is set to None if no snapshot exists for the time step.
        """
        if self._isnap == -1:
            istep = None
            isnap = -1
            # could be more efficient if do 0 and -1 then bisection
            # (but loose intermediate <- would probably use too much
            # memory for what it's worth if search algo is efficient)
            while (istep is None or istep < self.istep) and isnap < 99999:
                isnap += 1
                try:
                    istep = self.sdat.snaps[isnap].istep
                except KeyError:
                    pass
                # all intermediate istep could have their ._isnap to None
            if istep != self.istep:
                self._isnap = None
        return self._isnap



import pytest


@pytest.fixture(scope='session')
def repo_dir():
    import pathlib
    return pathlib.Path(__file__).parent.parent.resolve()


@pytest.fixture(scope='session',
                params=['ra-100000', 'annulus'])
def example_dir(request, repo_dir):
    return repo_dir / 'Examples' / request.param


@pytest.fixture(scope='module')
def sdat(example_dir):
    import stagpy.stagyydata
    return stagpy.stagyydata.StagyyData(example_dir)


@pytest.fixture(scope='module')
def step(sdat):
    return sdat.snaps[-1]



import re
import stagpy.commands


def test_info_cmd(capsys, example_dir):
    stagpy.conf.core.path = example_dir
    stagpy.commands.info_cmd()
    output = capsys.readouterr()
    expected = re.compile(
        r'^StagYY run in.*\n.* x .*\n\nStep.*, snapshot.*\n  t.*\n\n$',
        flags=re.DOTALL)
    assert expected.fullmatch(output.out)
    del stagpy.conf.core.path


def test_var_cmd(capsys):
    stagpy.commands.var_cmd()
    output = capsys.readouterr()
    expected = re.compile(
        r'field:\n.*\nrprof:\n.*\ntime:\n.*\nplates:\n.*$',
        flags=re.DOTALL)
    assert expected.fullmatch(output.out)


def test_version_cmd(capsys):
    stagpy.commands.version_cmd()
    output = capsys.readouterr()
    expected = 'stagpy version: {}\n'.format(stagpy.__version__)
    assert output.out == expected


def test_config_cmd(capsys):
    stagpy.commands.config_cmd()
    output = capsys.readouterr()
    expected = '(c|f): available only as CLI argument/in the config file'
    assert output.out.startswith(expected)



import stagpy.phyvars
import stagpy.time_series


def test_get_time_series(sdat):
    series, time, meta = sdat.tseries['Tmean']
    assert series.shape == time.shape
    assert meta == stagpy.phyvars.TIME['Tmean']


def test_get_time_series_extra(sdat):
    series, time, meta = sdat.tseries['dTdt']
    assert series.shape == time.shape == (sdat.tseries.time.size - 1,)
    assert isinstance(meta, stagpy.phyvars.Vart)


def test_compstat(sdat):
    variables = ['Tmean', 'vrms']
    stats = stagpy.time_series.compstat(sdat, *variables)
    assert list(stats.columns) == variables
    assert list(stats.index) == ['mean', 'rms']



"""Plate analysis."""

from contextlib import ExitStack

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import argrelextrema

from . import conf, error, field, misc, phyvars
from .stagyydata import StagyyData


def detect_plates_vzcheck(step, seuil_memz):
    """Detect plates and check with vz and plate size."""
    v_z = step.fields['v3'][0, :, :, 0]
    v_x = step.fields['v2'][0, :, :, 0]
    tcell = step.fields['T'][0, :, :, 0]
    n_z = step.nztot
    nphi = step.nptot  # -1? should be OK, ghost not included
    rcmb = step.geom.rcmb
    radius = step.geom.r_centers
    radiusgrid = step.geom.r_walls
    dphi = 1 / nphi

    # calculing temperature on the grid and vz_mean
    vz_mean = 0
    tgrid = np.zeros((nphi, n_z + 1))
    tgrid[:, 0] = 1
    for i_z in range(1, n_z):
        for phi in range(nphi):
            tgrid[phi, i_z] = (
                tcell[phi, i_z - 1] *
                (radiusgrid[i_z] - radius[i_z - 1]) + tcell[phi, i_z] *
                (-radiusgrid[i_z] + radius[i_z])) / (radius[i_z] -
                                                     radius[i_z - 1])
            vz_mean += abs(v_z[phi, i_z]) / (nphi * n_z)

    flux_c = n_z * [0]
    for i_z in range(1, n_z - 1):
        for phi in range(nphi):
            flux_c[i_z] += (tgrid[phi, i_z] - step.timeinfo.loc['Tmean']) * \
                v_z[phi, i_z] * radiusgrid[i_z] * dphi

    # checking stagnant lid
    if all(abs(flux_c[i_z]) <= np.max(flux_c) / 50
           for i_z in range(n_z - n_z // 20, n_z)):
        raise error.StagnantLidError(step.sdat)
    else:
        # verifying horizontal plate speed and closeness of plates
        dvphi = nphi * [0]
        dvx_thres = 16 * step.timeinfo.loc['vrms']

        for phi in range(0, nphi):
            dvphi[phi] = (v_x[phi, n_z - 1] -
                          v_x[phi - 1, n_z - 1]) / ((1 + rcmb) * dphi)
        limits = []
        for phi in range(0, nphi - nphi // 33):
            mark = all(abs(dvphi[i]) <= abs(dvphi[phi])
                       for i in range(phi - nphi // 33, phi + nphi // 33))
            if mark and abs(dvphi[phi]) >= dvx_thres:
                limits.append(phi)
        for phi in range(nphi - nphi // 33 + 1, nphi):
            mark = all(abs(dvphi[i]) <= abs(dvphi[phi])
                       for i in range(phi - nphi // 33 - nphi,
                                      phi + nphi // 33 - nphi))
            if mark and abs(dvphi[phi]) >= dvx_thres:
                limits.append(phi)

        # verifying vertical speed
        k = 0
        for i in range(len(limits)):
            vzm = 0
            phi = limits[i - k]
            for i_z in range(1 if phi == nphi - 1 else 0, n_z):
                vzm += (abs(v_z[phi, i_z]) +
                        abs(v_z[phi - 1, i_z]) +
                        abs(v_z[(phi + 1) % nphi, i_z])) / (n_z * 3)

            vz_thres = vz_mean * 0.1 + seuil_memz / 2 if seuil_memz else 0
            if vzm < vz_thres:
                limits.remove(phi)
                k += 1

    return limits, nphi, dvphi, vz_thres, v_x[:, n_z - 1]


def detect_plates(step, vrms_surface, fids, time):
    """Detect plates using derivative of horizontal velocity."""
    vphi = step.fields['v2'][0, :, :, 0]
    ph_coord = step.geom.p_centers

    if step.sdat.par['boundaries']['air_layer']:
        dsa = step.sdat.par['boundaries']['air_thickness']
        # we are a bit below the surface; should check if you are in the
        # thermal boundary layer
        indsurf = np.argmin(
            np.abs(1 - dsa - step.geom.r_centers + step.geom.rcmb)) - 4
    else:
        indsurf = -1

    vph2 = 0.5 * (vphi + np.roll(vphi, 1, 0))  # interpolate to the same phi
    # velocity derivation
    dvph2 = (np.diff(vph2[:, indsurf]) / (ph_coord[0] * 2.))

    io_surface(step.isnap, time, fids[6], dvph2)
    io_surface(step.isnap, time, fids[7], vph2[:-1, indsurf])

    # prepare stuff to find trenches and ridges
    myorder_trench = 15 if step.sdat.par['boundaries']['air_layer'] else 10
    myorder_ridge = 20  # threshold

    # finding trenches
    pom2 = np.copy(dvph2)
    if step.sdat.par['boundaries']['air_layer']:
        maskbigdvel = -30 * vrms_surface
    else:
        maskbigdvel = -10 * vrms_surface
    pom2[pom2 > maskbigdvel] = maskbigdvel
    argless_dv = argrelextrema(
        pom2, np.less, order=myorder_trench, mode='wrap')[0]
    trench = ph_coord[argless_dv]
    velocity_trench = vph2[argless_dv, indsurf]
    dv_trench = dvph2[argless_dv]

    # finding ridges
    pom2 = np.copy(dvph2)
    masksmalldvel = np.amax(dvph2) * 0.2
    pom2[pom2 < masksmalldvel] = masksmalldvel
    arggreat_dv = argrelextrema(
        pom2, np.greater, order=myorder_ridge, mode='wrap')[0]
    ridge = ph_coord[arggreat_dv]

    # elimination of ridges that are too close to trench
    argdel = []
    if trench.any() and ridge.any():
        for i, ridge_i in enumerate(ridge):
            mdistance = np.amin(abs(trench - ridge_i))
            if mdistance < 0.016:
                argdel.append(i)
        if argdel:
            print('deleting from ridge', trench, ridge[argdel])
            ridge = np.delete(ridge, np.array(argdel))
            arggreat_dv = np.delete(arggreat_dv, np.array(argdel))

    dv_ridge = dvph2[arggreat_dv]
    if 'age' in conf.plates:
        agefld = step.fields['age'][0, :, :, 0]
        age_surface = np.ma.masked_where(agefld[:, indsurf] < 0.00001,
                                         agefld[:, indsurf])
        age_surface_dim = age_surface * vrms_surface *\
            conf.scaling.ttransit / conf.scaling.yearins / 1.e6
        agetrench = age_surface_dim[argless_dv]  # age at the trench
    else:
        agetrench = np.zeros(len(argless_dv))

    # writing the output into a file, all time steps are in one file
    for itrench in np.arange(len(trench)):
        fids[0].write("%7.0f %11.7f %10.6f %9.2f %9.2f \n" % (
            step.isnap,
            step.geom.ti_ad,
            trench[itrench],
            velocity_trench[itrench],
            agetrench[itrench]
        ))

    return trench, ridge, agetrench, dv_trench, dv_ridge


def plot_plate_limits(axis, ridges, trenches, ymin, ymax):
    """Plot lines designating ridges and trenches."""
    for trench in trenches:
        axis.axvline(
            x=trench, ymin=ymin, ymax=ymax,
            color='red', ls='dashed', alpha=0.4)
    for ridge in ridges:
        axis.axvline(
            x=ridge, ymin=ymin, ymax=ymax,
            color='green', ls='dashed', alpha=0.4)
    axis.set_xlim(0, 2 * np.pi)
    axis.set_ylim(ymin, ymax)


def plot_plate_limits_field(axis, rcmb, ridges, trenches):
    """Plot arrows designating ridges and trenches in 2D field plots."""
    for trench in trenches:
        xxd = (rcmb + 1.02) * np.cos(trench)  # arrow begin
        yyd = (rcmb + 1.02) * np.sin(trench)  # arrow begin
        xxt = (rcmb + 1.35) * np.cos(trench)  # arrow end
        yyt = (rcmb + 1.35) * np.sin(trench)  # arrow end
        axis.annotate('', xy=(xxd, yyd), xytext=(xxt, yyt),
                      arrowprops=dict(facecolor='red', shrink=0.05))
    for ridge in ridges:
        xxd = (rcmb + 1.02) * np.cos(ridge)
        yyd = (rcmb + 1.02) * np.sin(ridge)
        xxt = (rcmb + 1.35) * np.cos(ridge)
        yyt = (rcmb + 1.35) * np.sin(ridge)
        axis.annotate('', xy=(xxd, yyd), xytext=(xxt, yyt),
                      arrowprops=dict(facecolor='green', shrink=0.05))


def plot_plates(step, time, vrms_surface, trench, ridge, agetrench,
                topo, fids):
    """Handle plotting stuff."""
    vphi = step.fields['v2'][0, :, :, 0]
    tempfld = step.fields['T'][0, :, :, 0]
    concfld = step.fields['c'][0, :, :, 0]
    timestep = step.isnap

    if step.sdat.par['boundaries']['air_layer']:
        dsa = step.sdat.par['boundaries']['air_thickness']
        # we are a bit below the surface; delete "-some number"
        # to be just below
        # the surface (that is considered plane here); should check if you are
        # in the thermal boundary layer
        indsurf = np.argmin(
            np.abs(1 - dsa - step.geom.r_centers + step.geom.rcmb)) - 4
        # depth to detect the continents
        indcont = np.argmin(
            np.abs(1 - dsa - step.geom.r_centers + step.geom.rcmb)) - 10
    else:
        indsurf = -1
        indcont = -1  # depth to detect continents

    if step.sdat.par['boundaries']['air_layer'] and\
       not step.sdat.par['continents']['proterozoic_belts']:
        continents = np.ma.masked_where(
            np.logical_or(concfld[:-1, indcont] < 3,
                          concfld[:-1, indcont] > 4),
            concfld[:-1, indcont])
    elif (step.sdat.par['boundaries']['air_layer'] and
          step.sdat.par['continents']['proterozoic_belts']):
        continents = np.ma.masked_where(
            np.logical_or(concfld[:-1, indcont] < 3,
                          concfld[:-1, indcont] > 5),
            concfld[:-1, indcont])
    elif step.sdat.par['tracersin']['tracers_weakcrust']:
        continents = np.ma.masked_where(
            concfld[:-1, indcont] < 3, concfld[:-1, indcont])
    else:
        continents = np.ma.masked_where(
            concfld[:-1, indcont] < 2, concfld[:-1, indcont])

    # masked array, only continents are true
    continentsall = continents / continents

    ph_coord = step.geom.p_centers

    # velocity
    vph2 = 0.5 * (vphi + np.roll(vphi, 1, 0))  # interpolate to the same phi
    dvph2 = (np.diff(vph2[:, indsurf]) / (ph_coord[0] * 2.))

    # plotting
    fig0, (ax1, ax2, ax3) = plt.subplots(3, 1, sharex=True, figsize=(12, 8))
    ax1.plot(ph_coord[:-1], concfld[:-1, indsurf], color='g', label='Conc')
    ax2.plot(ph_coord[:-1], tempfld[:-1, indsurf], color='k', label='Temp')
    ax3.plot(ph_coord[:-1], vph2[:-1, indsurf], label='Vel')

    ax1.fill_between(
        ph_coord[:-1], continents, 1., facecolor='#8B6914', alpha=0.2)
    ax2.fill_between(
        ph_coord[:-1], continentsall, 0., facecolor='#8B6914', alpha=0.2)

    tempmin = step.sdat.par['boundaries']['topT_val'] * 0.9\
        if step.sdat.par['boundaries']['topT_mode'] == 'iso' else 0.0
    tempmax = step.sdat.par['boundaries']['botT_val'] * 0.35\
        if step.sdat.par['boundaries']['botT_mode'] == 'iso' else 0.8

    ax2.set_ylim(tempmin, tempmax)
    ax3.fill_between(
        ph_coord[:-1], continentsall * round(1.5 * np.amax(dvph2), 1),
        round(np.amin(dvph2) * 1.1, 1), facecolor='#8B6914', alpha=0.2)
    ax3.set_ylim(conf.plates.vmin, conf.plates.vmax)

    ax1.set_ylabel("Concentration")
    ax2.set_ylabel("Temperature")
    ax3.set_ylabel("Velocity")
    ax1.set_title(timestep)
    ax1.text(0.95, 1.07, str(round(time, 0)) + ' My',
             transform=ax1.transAxes)
    ax1.text(0.01, 1.07, str(round(step.geom.ti_ad, 8)),
             transform=ax1.transAxes)

    plot_plate_limits(ax3, ridge, trench, conf.plates.vmin,
                      conf.plates.vmax)

    misc.saveplot(fig0, 'sveltempconc', timestep)

    # plotting velocity and velocity derivative
    fig0, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(12, 8))
    ax1.plot(ph_coord[:-1], vph2[:-1, indsurf], label='Vel')
    ax1.axhline(y=0, xmin=0, xmax=2 * np.pi,
                color='black', ls='solid', alpha=0.2)
    ax1.set_ylabel("Velocity")
    ax1.text(0.95, 1.07, str(round(time, 0)) + ' My',
             transform=ax1.transAxes)
    ax1.text(0.01, 1.07, str(round(step.geom.ti_ad, 8)),
             transform=ax1.transAxes)
    ax2.plot(ph_coord[:-1] + ph_coord[0], dvph2, color='k', label='dv')
    ax2.set_ylabel("dv")

    plot_plate_limits(ax1, ridge, trench, conf.plates.vmin,
                      conf.plates.vmax)
    plot_plate_limits(ax2, ridge, trench, conf.plates.dvmin,
                      conf.plates.dvmax)
    ax1.set_xlim(0, 2 * np.pi)
    ax1.set_title(timestep)

    ax1.fill_between(
        ph_coord[:-1], continentsall * conf.plates.vmin, conf.plates.vmax,
        facecolor='#8b6914', alpha=0.2)
    ax1.set_ylim(conf.plates.vmin, conf.plates.vmax)
    ax2.fill_between(
        ph_coord[:-1], continentsall * conf.plates.dvmin,
        conf.plates.dvmax, facecolor='#8b6914', alpha=0.2)
    ax2.set_ylim(conf.plates.dvmin, conf.plates.dvmax)

    misc.saveplot(fig0, 'sveldvel', timestep)

    # plotting velocity and second invariant of stress
    if 'str' in conf.plates.plot:
        stressfld = step.fields['sII'][0, :, :, 0]
        fig0, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(12, 8))
        ax1.plot(ph_coord[:-1], vph2[:-1, indsurf], label='Vel')
        ax1.axhline(y=0, xmin=0, xmax=2 * np.pi,
                    color='black', ls='solid', alpha=0.2)
        ax1.set_ylabel("Velocity")
        ax1.text(0.95, 1.07, str(round(time, 0)) + ' My',
                 transform=ax1.transAxes)
        ax1.text(0.01, 1.07, str(round(step.geom.ti_ad, 8)),
                 transform=ax1.transAxes)
        ax2.plot(ph_coord[:-1],
                 stressfld[:-1, indsurf] * step.sdat.scales.stress / 1.e6,
                 color='k', label='Stress')
        ax2.set_ylim(conf.plates.stressmin, conf.plates.stressmax)
        ax2.set_ylabel("Stress [MPa]")

        plot_plate_limits(ax1, ridge, trench,
                          conf.plates.vmin, conf.plates.vmax)
        plot_plate_limits(ax2, ridge, trench,
                          conf.plates.stressmin, conf.plates.stressmax)
        ax1.set_xlim(0, 2 * np.pi)
        ax1.set_title(timestep)

        ax1.fill_between(
            ph_coord[:-1], continentsall * conf.plates.vmin,
            conf.plates.vmax, facecolor='#8B6914', alpha=0.2)
        ax1.set_ylim(conf.plates.vmin, conf.plates.vmax)
        ax2.fill_between(
            ph_coord[:-1], continentsall * conf.plates.dvmin,
            conf.plates.dvmax,
            facecolor='#8B6914', alpha=0.2)

        misc.saveplot(fig0, 'svelstress', timestep)

    # plotting velocity
    fig1, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(12, 8))
    ax1.plot(ph_coord[:-1], vph2[:-1, indsurf], label='Vel')
    ax1.axhline(y=0, xmin=0, xmax=2 * np.pi,
                color='black', ls='solid', alpha=0.2)
    ax1.set_ylim(conf.plates.vmin, conf.plates.vmax)
    ax1.set_ylabel("Velocity")
    ax1.text(0.95, 1.07, str(round(time, 0)) + ' My',
             transform=ax1.transAxes)
    plot_plate_limits(ax1, ridge, trench, conf.plates.vmin,
                      conf.plates.vmax)

    # plotting velocity and age at surface
    if 'age' in conf.plates.plot:
        agefld = step.fields['age'][0, :, :, 0]
        age_surface = np.ma.masked_where(
            agefld[:, indsurf] < 0.00001, agefld[:, indsurf])
        age_surface_dim = (age_surface * vrms_surface * conf.scaling.ttransit /
                           conf.scaling.yearins / 1.e6)

        fig2, (ax3, ax4) = plt.subplots(2, 1, sharex=True, figsize=(12, 8))
        ax3.plot(ph_coord[:-1], vph2[:-1, indsurf], label='Vel')
        ax3.axhline(
            y=0, xmin=0, xmax=2 * np.pi,
            color='black', ls='solid', alpha=0.2)
        ax3.set_ylim(conf.plates.vmin, conf.plates.vmax)
        ax3.set_ylabel("Velocity")
        ax3.text(0.95, 1.07, str(round(time, 0)) + ' My',
                 transform=ax3.transAxes)
        ax3.fill_between(
            ph_coord[:-1], continentsall * conf.plates.vmax,
            conf.plates.vmin, facecolor='#8B6914', alpha=0.2)
        plot_plate_limits(ax3, ridge, trench,
                          conf.plates.vmin, conf.plates.vmax)

    times_subd = []
    age_subd = []
    distance_subd = []
    ph_trench_subd = []
    ph_cont_subd = []
    if step.sdat.par['switches']['cont_tracers']:
        for i, trench_i in enumerate(trench):
            # detection of the distance in between subduction and continent
            ph_coord_noend = ph_coord[:-1]
            angdistance1 = abs(ph_coord_noend[continentsall == 1] - trench_i)
            angdistance2 = 2. * np.pi - angdistance1
            angdistance = np.minimum(angdistance1, angdistance2)
            distancecont = min(angdistance)
            argdistancecont = np.argmin(angdistance)
            continentpos = ph_coord_noend[continentsall == 1][argdistancecont]

            ph_trench_subd.append(trench_i)
            age_subd.append(agetrench[i])
            ph_cont_subd.append(continentpos)
            distance_subd.append(distancecont)
            times_subd.append(step.geom.ti_ad)

            if angdistance1[argdistancecont] < angdistance2[argdistancecont]:
                if continentpos - trench_i < 0:  # continent is on the left
                    distancecont = - distancecont
                ax1.annotate('', xy=(trench_i + distancecont, 2000),
                             xycoords='data', xytext=(trench_i, 2000),
                             textcoords='data',
                             arrowprops=dict(arrowstyle="->", lw="2",
                                             shrinkA=0, shrinkB=0))
            else:  # distance over boundary
                xy_anot, xy_text = 0, 2 * np.pi
                if continentpos - trench_i < 0:
                    xy_anot, xy_text = xy_text, xy_anot
                ax1.annotate('', xy=(xy_anot, 2000),
                             xycoords='data', xytext=(trench_i, 2000),
                             textcoords='data',
                             arrowprops=dict(arrowstyle="-", lw="2",
                                             shrinkA=0, shrinkB=0))
                ax1.annotate('', xy=(continentpos, 2000),
                             xycoords='data', xytext=(xy_text, 2000),
                             textcoords='data',
                             arrowprops=dict(arrowstyle="->", lw="2",
                                             shrinkA=0, shrinkB=0))

    ax1.fill_between(
        ph_coord[:-1], continentsall * conf.plates.vmin,
        conf.plates.vmax, facecolor='#8B6914', alpha=0.2)
    ax2.set_ylabel("Topography [km]")
    ax2.axhline(y=0, xmin=0, xmax=2 * np.pi,
                color='black', ls='solid', alpha=0.2)
    ax2.plot(topo[:, 0],
             topo[:, 1] * step.sdat.scales.length / 1.e3,
             color='black')
    ax2.set_xlim(0, 2 * np.pi)
    ax2.set_ylim(conf.plates.topomin, conf.plates.topomax)
    ax2.fill_between(
        ph_coord[:-1], continentsall * conf.plates.topomax,
        conf.plates.topomin, facecolor='#8B6914', alpha=0.2)
    plot_plate_limits(ax2, ridge, trench, conf.plates.topomin,
                      conf.plates.topomax)
    ax1.set_title(timestep)
    misc.saveplot(fig1, 'sveltopo', timestep)

    if 'age' in conf.plates.plot:
        ax4.set_ylabel("Seafloor age [My]")
        # in dimensions
        ax4.plot(ph_coord[:-1], age_surface_dim[:-1], color='black')
        ax4.set_xlim(0, 2 * np.pi)
        ax4.fill_between(
            ph_coord[:-1], continentsall * conf.plates.agemax,
            conf.plates.agemin, facecolor='#8B6914', alpha=0.2)
        ax4.set_ylim(conf.plates.agemin, conf.plates.agemax)
        plot_plate_limits(ax4, ridge, trench, conf.plates.agemin,
                          conf.plates.agemax)
        ax3.set_title(timestep)
        misc.saveplot(fig2, 'svelage', timestep)

    # writing the output into a file, all time steps are in one file
    for isubd in np.arange(len(distance_subd)):
        fids[1].write("%6.0f %11.7f %11.3f %10.6f %10.6f %10.6f %11.3f\n" % (
            timestep,
            times_subd[isubd],
            time,
            distance_subd[isubd],
            ph_trench_subd[isubd],
            ph_cont_subd[isubd],
            age_subd[isubd],
        ))


def io_surface(timestep, time, fid, fld):
    """Output surface files."""
    fid.write(f"{timestep} {time}")
    fid.writelines(["%10.2e" % item for item in fld[:]])
    fid.writelines(["\n"])


def lithospheric_stress(step, trench, ridge, time):
    """Calculate stress in the lithosphere."""
    timestep = step.isnap
    base_lith = step.geom.rcmb + 1 - 0.105

    stressfld = step.fields['sII'][0, :, :, 0]
    r_centers = np.outer(np.ones(stressfld.shape[0]), step.geom.r_centers)
    stressfld = np.ma.masked_where(r_centers < base_lith, stressfld)

    # stress integration in the lithosphere
    dzm = (step.geom.r_centers[1:] - step.geom.r_centers[:-1])
    stress_lith = np.sum((stressfld[:, 1:] * dzm.T), axis=1)
    ph_coord = step.geom.p_centers  # probably doesn't need alias

    # plot stress in the lithosphere
    fig, axis, _, _ = field.plot_scalar(step, 'sII', stressfld,
                                        cmap='plasma_r', vmin=0, vmax=300)
    # Annotation with time and step
    axis.text(1., 0.9, str(round(time, 0)) + ' My', transform=axis.transAxes)
    axis.text(1., 0.1, str(timestep), transform=axis.transAxes)
    misc.saveplot(fig, 'lith', timestep)

    # velocity
    vphi = step.fields['v2'][0, :, :, 0]
    vph2 = 0.5 * (vphi + np.roll(vphi, 1, 0))  # interpolate to the same phi

    # position of continents
    concfld = step.fields['c'][0, :, :, 0]
    if step.sdat.par['boundaries']['air_layer']:
        # we are a bit below the surface; delete "-some number"
        # to be just below
        dsa = step.sdat.par['boundaries']['air_thickness']
        # depth to detect the continents
        indcont = np.argmin(
            np.abs(1 - dsa - step.geom.r_centers + step.geom.rcmb)) - 10
    else:
        # depth to detect continents
        indcont = -1
    if step.sdat.par['boundaries']['air_layer'] and\
            not step.sdat.par['continents']['proterozoic_belts']:
        continents = np.ma.masked_where(
            np.logical_or(concfld[:-1, indcont] < 3,
                          concfld[:-1, indcont] > 4),
            concfld[:-1, indcont])
    elif step.sdat.par['boundaries']['air_layer'] and\
            step.sdat.par['continents']['proterozoic_belts']:
        continents = np.ma.masked_where(
            np.logical_or(concfld[:-1, indcont] < 3,
                          concfld[:-1, indcont] > 5),
            concfld[:-1, indcont])
    elif step.sdat.par['tracersin']['tracers_weakcrust']:
        continents = np.ma.masked_where(
            concfld[:-1, indcont] < 3, concfld[:-1, indcont])
    else:
        continents = np.ma.masked_where(
            concfld[:-1, indcont] < 2, concfld[:-1, indcont])

    # masked array, only continents are true
    continentsall = continents / continents

    # plot integrated stress in the lithosphere
    fig0, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(12, 8))
    ax1.plot(ph_coord[:-1], vph2[:-1, -1], label='Vel')
    ax1.axhline(y=0, xmin=0, xmax=2 * np.pi,
                color='black', ls='solid', alpha=0.2)
    ax1.set_ylabel("Velocity")
    ax1.text(0.95, 1.07, str(round(time, 0)) + ' My',
             transform=ax1.transAxes)
    ax1.text(0.01, 1.07, str(round(step.geom.ti_ad, 8)),
             transform=ax1.transAxes)

    intstr_scale = step.sdat.scales.stress * step.sdat.scales.length / 1.e12
    ax2.plot(ph_coord, stress_lith * intstr_scale, color='k', label='Stress')
    ax2.set_ylabel(r"Integrated stress [$TN\,m^{-1}$]")

    plot_plate_limits(ax1, ridge, trench, conf.plates.vmin,
                      conf.plates.vmax)
    plot_plate_limits(ax2, ridge, trench, conf.plates.stressmin,
                      conf.plates.lstressmax)
    ax1.set_xlim(0, 2 * np.pi)
    ax1.set_title(timestep)

    ax1.fill_between(
        ph_coord[:-1], continentsall * conf.plates.vmin,
        conf.plates.vmax, facecolor='#8b6914', alpha=0.2)
    ax1.set_ylim(conf.plates.vmin, conf.plates.vmax)
    ax2.fill_between(
        ph_coord[:-1], continentsall * conf.plates.stressmin,
        conf.plates.lstressmax, facecolor='#8b6914', alpha=0.2)
    ax2.set_ylim(conf.plates.stressmin, conf.plates.lstressmax)

    misc.saveplot(fig0, 'svelslith', timestep)


def set_of_vars(arg_plot):
    """Build set of needed variables.

    Args:
        arg_plot (str): string with variable names separated with ``,``.
    Returns:
        set of str: set of variables.
    """
    return set(var for var in arg_plot.split(',') if var in phyvars.PLATES)


def main_plates(sdat):
    """Plot several plates information."""
    # calculating averaged horizontal surface velocity
    # needed for redimensionalisation
    rlast = sdat.snaps[-1].rprofs
    nprof = 0
    uprof_averaged = np.zeros_like(rlast['vhrms'].values)
    for step in sdat.walk.filter(rprofs=True):
        uprof_averaged += step.rprofs['vhrms'].values
        nprof += 1
    uprof_averaged /= nprof
    radius = rlast['r'].values
    if sdat.par['boundaries']['air_layer']:
        dsa = sdat.par['boundaries']['air_thickness']
        isurf = np.argmin(abs(radius - radius[-1] + dsa))
        vrms_surface = uprof_averaged.iloc[isurf]
        isurf = np.argmin(abs((1 - dsa) - radius))
        isurf -= 4  # why different isurf for the rest?
    else:
        isurf = -1
        vrms_surface = uprof_averaged.iloc[isurf]

    # determine names of files
    fnames = ['plate_velocity', 'distance_subd', 'continents',
              'flux', 'topography', 'age', 'velderiv', 'velocity']
    fnames = [f'plates_{stem}_{sdat.walk.stepstr}' for stem in fnames]
    with ExitStack() as stack:
        fids = [stack.enter_context(open(fname)) for fname in fnames]
        fids[0].write('#  it  time  ph_trench vel_trench age_trench\n')
        fids[1].write('#  it      time   time [My]   distance     '
                      'ph_trench     ph_cont  age_trench [My]\n')

        for step in sdat.walk.filter(fields=['T']):
            # could check other fields too
            timestep = step.isnap
            print('Treating snapshot', timestep)

            rcmb = step.geom.rcmb
            # topography
            fname = sdat.filename('sc', timestep=timestep, suffix='.dat')
            topo = np.genfromtxt(str(fname))
            # rescaling topography!
            if sdat.par['boundaries']['air_layer']:
                topo[:, 1] = topo[:, 1] / (1. - dsa)

            time = step.geom.ti_ad * vrms_surface *\
                conf.scaling.ttransit / conf.scaling.yearins / 1.e6
            trenches, ridges, agetrenches, _, _ =\
                detect_plates(step, vrms_surface, fids, time)
            plot_plates(step, time, vrms_surface, trenches, ridges,
                        agetrenches, topo, fids)

            # prepare for continent plotting
            concfld = step.fields['c'][0, :, :, 0]
            continentsfld = np.ma.masked_where(
                concfld < 3, concfld)  # plotting continents, to-do
            continentsfld = continentsfld / continentsfld

            temp = step.fields['T'][0, :, :, 0]
            tgrad = (temp[:, isurf - 1] - temp[:, isurf]) /\
                (step.geom.r_centers[isurf] - step.geom.r_centers[isurf - 1])

            io_surface(timestep, time, fids[2], concfld[:-1, isurf])
            io_surface(timestep, time, fids[3], tgrad)
            io_surface(timestep, time, fids[4], topo[:, 1])
            if 'age' in conf.plates.plot:
                io_surface(timestep, time, fids[5],
                           step.fields['age'][0, :, isurf, 0])

            # plot viscosity field with position of trenches and ridges
            etamin, _ = sdat.scale(1e-2, 'Pa')
            etamax, _ = sdat.scale(sdat.par['viscosity']['eta_max'], 'Pa')
            fig, axis, _, _ = field.plot_scalar(step, 'eta',
                                                vmin=etamin, vmax=etamax)

            # plotting continents
            cbar = conf.field.colorbar
            conf.field.colorbar = False
            field.plot_scalar(step, 'c', continentsfld, axis,
                              cmap='cool_r', vmin=0, vmax=0)
            cmap2 = plt.cm.ocean
            cmap2.set_over('m')
            conf.field.colorbar = cbar

            # plotting velocity vectors
            field.plot_vec(axis, step, 'v')

            # Annotation with time and step
            axis.text(1., 0.9, str(round(time, 0)) + ' My',
                      transform=axis.transAxes)
            axis.text(1., 0.1, str(timestep),
                      transform=axis.transAxes)

            # Put arrow where ridges and trenches are
            plot_plate_limits_field(axis, rcmb, ridges, trenches)

            misc.saveplot(fig, 'eta', timestep, close=False)

            # Zoom
            if conf.plates.zoom is not None:
                if not 0 <= conf.plates.zoom <= 360:
                    raise error.InvalidZoomError(conf.plates.zoom)
                if 45 < conf.plates.zoom <= 135:
                    ladd, radd, uadd, dadd = 0.8, 0.8, 0.05, 0.1
                elif 135 < conf.plates.zoom <= 225:
                    ladd, radd, uadd, dadd = 0.05, 0.1, 0.8, 0.8
                elif 225 < conf.plates.zoom <= 315:
                    ladd, radd, uadd, dadd = 0.8, 0.8, 0.1, 0.05
                else:  # >315 or <=45
                    ladd, radd, uadd, dadd = 0.1, 0.05, 0.8, 0.8
                xzoom = (rcmb + 1) * np.cos(np.radians(conf.plates.zoom))
                yzoom = (rcmb + 1) * np.sin(np.radians(conf.plates.zoom))
                axis.set_xlim(xzoom - ladd, xzoom + radd)
                axis.set_ylim(yzoom - dadd, yzoom + uadd)
                misc.saveplot(fig, 'etazoom', timestep, close=False)
            plt.close(fig)

            # plot stress field with position of trenches and ridges
            if 'str' in conf.plates.plot:
                fig, axis, _, _ = field.plot_scalar(step, 'sII',
                                                    vmin=0, vmax=300)

                # Annotation with time and step
                axis.text(1., 0.9, str(round(time, 0)) + ' My',
                          transform=axis.transAxes)
                axis.text(1., 0.1, str(timestep),
                          transform=axis.transAxes)

                # Put arrow where ridges and trenches are
                plot_plate_limits_field(axis, rcmb, ridges, trenches)

                misc.saveplot(fig, 's', timestep, close=False)

                # Zoom
                if conf.plates.zoom is not None:
                    axis.set_xlim(xzoom - ladd, xzoom + radd)
                    axis.set_ylim(yzoom - dadd, yzoom + uadd)
                    misc.saveplot(fig, 'szoom', timestep, close=False)
                plt.close(fig)

                # calculate stresses in the lithosphere
                lithospheric_stress(step, trenches, ridges, time)

            # plotting the principal deviatoric stress field
            if 'sx' in conf.plates.plot:
                fig, axis, _, _ = field.plot_scalar(step, 'sII',
                                                    alpha=0.1)

                # plotting continents
                cbar = conf.field.colorbar
                conf.field.colorbar = False
                field.plot_scalar(step, 'c', continentsfld, axis,
                                  cmap='cool_r', vmin=0, vmax=0)
                cmap2 = plt.cm.ocean
                cmap2.set_over('m')
                conf.field.colorbar = cbar

                # plotting principal deviatoric stress
                field.plot_vec(axis, step, 'sx')

                # Annotation with time and step
                axis.text(1., 0.9, str(round(time, 0)) + ' My',
                          transform=axis.transAxes)
                axis.text(1., 0.1, str(timestep),
                          transform=axis.transAxes)

                # Put arrow where ridges and trenches are
                plot_plate_limits_field(axis, rcmb, ridges, trenches)

                misc.saveplot(fig, 'sx', timestep)


def cmd():
    """Implementation of plates subcommand.

    Other Parameters:
        conf.plates
        conf.scaling
        conf.plot
        conf.core
    """
    sdat = StagyyData()
    conf.plates.plot = set_of_vars(conf.plates.plot)
    if not conf.plates.vzcheck:
        conf.scaling.dimensional = True
        conf.scaling.factors['Pa'] = 'M'
        main_plates(sdat)
    else:
        seuil_memz = 0
        nb_plates = []
        time = []
        ch2o = []
        istart, iend = None, None

        for step in sdat.walk.filter(fields=['T']):
            # could check other fields too
            if conf.plates.timeprofile:
                time.append(step.timeinfo.loc['t'])
                ch2o.append(step.timeinfo.loc[0])
            istart = step.isnap if istart is None else istart
            iend = step.isnap
            limits, nphi, dvphi, seuil_memz, vphi_surf =\
                detect_plates_vzcheck(step, seuil_memz)
            water_profile = np.mean(step.fields['wtr'][0, :, :, 0], axis=0)
            limits.sort()
            sizeplates = [limits[0] + nphi - limits[-1]]
            for lim in range(1, len(limits)):
                sizeplates.append(limits[lim] - limits[lim - 1])
            lim = len(limits) * [max(dvphi)]
            fig = plt.figure()
            plt.subplot(221)
            plt.axis([0, nphi,
                      np.min(vphi_surf) * 1.2, np.max(vphi_surf) * 1.2])
            plt.plot(vphi_surf)
            plt.subplot(223)
            plt.axis(
                [0, nphi,
                 np.min(dvphi) * 1.2, np.max(dvphi) * 1.2])
            plt.plot(dvphi)
            plt.scatter(limits, lim, color='red')
            plt.subplot(222)
            plt.hist(sizeplates, 10, (0, nphi / 2))
            plt.subplot(224)
            plt.plot(water_profile)
            misc.saveplot(fig, 'plates', step.isnap)

            nb_plates.append(len(limits))

        if conf.plates.timeprofile:
            for i in range(2, len(nb_plates) - 3):
                nb_plates[i] = (nb_plates[i - 2] + nb_plates[i - 1] +
                                nb_plates[i] + nb_plates[i + 1] +
                                nb_plates[i + 2]) / 5
            figt = plt.figure()
            plt.subplot(121)
            plt.axis([time[0], time[-1], 0, np.max(nb_plates)])
            plt.plot(time, nb_plates)
            plt.subplot(122)
            plt.plot(time, ch2o)
            misc.saveplot(figt, f'plates_{istart}_{iend}')



import pathlib
import pytest
import stagpy.error
import stagpy.parfile


@pytest.fixture
def par_nml(example_dir):
    return stagpy.parfile.readpar(example_dir / 'par', example_dir)


def test_section_present(par_nml):
    for section in stagpy.parfile.PAR_DEFAULT.keys():
        assert section in par_nml


def test_section_case_insensitive(par_nml):
    return par_nml['SwitcHes'] is par_nml['switches']


def test_option_case_insensitive(par_nml):
    return par_nml['refstate']['Ra0'] is par_nml['refstate']['ra0']


def test_no_par_file():
    invalid = pathlib.Path('dummyinvalidpar')
    with pytest.raises(stagpy.error.NoParFileError) as err:
        stagpy.parfile.readpar(invalid, invalid)
    assert err.value.parfile == invalid



"""Plot reference state profiles."""

import matplotlib.pyplot as plt

from . import conf, misc
from .phyvars import REFSTATE
from .stagyydata import StagyyData


def plot_ref(sdat, var):
    """Plot one reference state.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
        var (str): refstate variable, a key of :data:`stagpy.phyvars.REFSTATE`.
    """
    fig, axis = plt.subplots()
    adbts = sdat.refstate.adiabats
    if len(adbts) > 2:
        for iad, adia in enumerate(adbts[:-1], 1):
            axis.plot(adia[var], adia['z'],
                      conf.refstate.style,
                      label=f'System {iad}')
    axis.plot(adbts[-1][var], adbts[-1]['z'],
              conf.refstate.style, color='k',
              label='Combined profile')
    if var == 'Tcond':
        axis.set_xscale('log')
    axis.set_xlabel(REFSTATE[var].description)
    axis.set_ylabel('z Position')
    if len(adbts) > 2:
        axis.legend()
    misc.saveplot(fig, f'refstate_{var}')


def cmd():
    """Implementation of refstate subcommand.

    Other Parameters:
        conf.core
        conf.plot
    """
    sdat = StagyyData()
    if sdat.refstate.adiabats is None:
        return

    lov = conf.refstate.plot.split(',')
    if not lov:
        return
    for var in lov:
        plot_ref(sdat, var)



import stagpy
import stagpy.misc


def test_out_name_conf():
    oname = 'something_fancy'
    stagpy.conf.core.outname = oname
    stem = 'teapot'
    assert stagpy.misc.out_name(stem) == oname + '_' + stem
    del stagpy.conf.core.outname


def test_out_name_number():
    assert stagpy.misc.out_name('T', 123) == 'stagpy_T00123'


def test_baredoc():
    """
       Badly formatted docstring .. .

    With some content.

    """
    expected = 'Badly formatted docstring'
    assert stagpy.misc.baredoc(test_baredoc) == expected


def test_list_of_vars():
    expected = [[['a', 'b'], ['c', 'd', 'e']], [['f', 'g'], ['h']]]
    assert stagpy.misc.list_of_vars('a,b..c,d,,,e-f,g.h-,..,-') == expected



import pathlib
from stagpy import stagyyparsers as prs


def test_time_series_prs(sdat):
    names = ['aa', 'bb', 'cc']
    data = prs.time_series(sdat.filename('time.dat'), list(names))
    assert (data.columns[:3] == names).all()
    assert (data.columns[3:] == list(range(data.shape[1] - 3))).all()


def test_time_series_invalid_prs():
    assert prs.time_series(pathlib.Path('dummy'), []) is None


def test_rprof_prs(sdat):
    names = ['aa', 'bb', 'cc']
    data, time = prs.rprof(sdat.filename('rprof.dat'), list(names))
    assert all((df.columns[:3] == names).all() for df in data.values())
    assert all((df.columns[3:] == list(range(df.shape[1] - 3))).all()
               for df in data.values())


def test_rprof_invalid_prs():
    assert prs.rprof(pathlib.Path('dummy'), []) == ({}, None)


def test_fields_prs(sdat):
    hdr, flds = prs.fields(sdat.filename('t', len(sdat.snaps) - 1))
    assert flds.shape[0] == 1
    assert flds.shape[4] == 1
    assert flds.shape[1:4] == tuple(hdr['nts'])


def test_fields_header_only_prs(sdat):
    hdr = prs.fields(sdat.filename('t', len(sdat.snaps) - 1), only_header=True)
    assert hdr['nts'].shape == (3,)


def test_fields_istep_only_prs(sdat):
    istep = prs.fields(sdat.filename('t', len(sdat.snaps) - 1),
                       only_istep=True)
    assert istep == sdat.snaps[-1].istep


def test_fields_invalid_prs():
    assert prs.fields(pathlib.Path('dummy')) is None



"""Lists of physical variables made available by StagPy.

They are organized by kind of variables (field, profiles, and time series).
EXTRA lists group variables that are not directly output by StagYY and need to
be computed from other variables.
"""

from collections import OrderedDict, namedtuple
from operator import attrgetter

from . import processing

Varf = namedtuple('Varf', ['description', 'dim'])
FIELD = OrderedDict((
    ('T', Varf('Temperature', 'K')),
    ('v1', Varf('x Velocity', 'm/s')),
    ('v2', Varf('y Velocity', 'm/s')),
    ('v3', Varf('z Velocity', 'm/s')),
    ('p', Varf('Pressure', 'Pa')),
    ('eta', Varf('Viscosity', 'Pa.s')),
    ('rho', Varf('Density', 'kg/m3')),
    ('rho4rhs', Varf('Density term in RHS', 'kg/m3')),
    ('trarho', Varf('Density from tracer mass', 'kg/m3')),
    ('sII', Varf('Second invariant of stress tensor', 'Pa')),
    ('sx1', Varf('1st comp. of principal stress eigenvector', 'Pa')),
    ('sx2', Varf('2nd comp. of principal stress eigenvector', 'Pa')),
    ('sx3', Varf('3rd comp. of principal stress eigenvector', 'Pa')),
    ('s1val', Varf('Principal stress eigenvalue', 'Pa')),
    ('edot', Varf('Strain rate', '1/s')),
    ('Tcond1', Varf('x Conductivity', 'W/(m.K)')),
    ('Tcond2', Varf('y Conductivity', 'W/(m.K)')),
    ('Tcond3', Varf('z Conductivity', 'W/(m.K)')),
    ('c', Varf('Composition', '1')),
    ('cFe', Varf('FeO content', '1')),
    ('hpe', Varf('HPE content', '1')),
    ('wtr', Varf('Water concentration', '1')),
    ('age', Varf('Age', 's')),
    ('contID', Varf('ID of continents', '1')),
    ('rs1', Varf('x Momentum residue', '1')),
    ('rs2', Varf('y Momentum residue', '1')),
    ('rs3', Varf('z Momentum residue', '1')),
    ('rsc', Varf('Continuity residue', '1')),
    ('prim', Varf('Primordial layer', '1')),
))

FIELD_EXTRA = OrderedDict((
    ('stream', Varf(processing.stream_function, 'm2/s')),
))

FIELD_FILES = OrderedDict((
    ('t', ['T']),
    ('vp', ['v1', 'v2', 'v3', 'p']),
    ('c', ['c']),
    ('eta', ['eta']),
    ('rho', ['rho']),
    ('hpe', ['hpe']),
    ('wtr', ['wtr']),
    ('age', ['age']),
    ('nrc', ['contID']),
    ('rs', ['rs1', 'rs2', 'rs3', 'rsc']),
    ('str', ['sII']),
    ('sx', ['sx1', 'sx2', 'sx3', 's1val']),
    ('ed', ['edot']),
    ('tcond', ['Tcond1', 'Tcond2', 'Tcond3']),
    ('prm', ['prim']),
))

FIELD_FILES_H5 = OrderedDict((
    ('Temperature', ['T']),
    ('Velocity', ['v1', 'v2', 'v3']),
    ('Dynamic_Pressure', ['p']),
    # Depending on the version of StagYY and whether total pressure is used,
    # the dynamic pressure might be written in this file instead.  It will be
    # used as fallback by _step._Fields._get_raw_data.
    ('Pressure', ['p']),
    ('Composition', ['c']),
    ('IronContent', ['cFe']),
    ('HPE', ['hpe']),
    ('Viscosity', ['eta']),
    ('Density', ['rho']),
    ('Density4rhs', ['rho4rhs']),
    ('TraBasedDensity', ['trarho']),
    ('water', ['wtr']),
    ('Age', ['age']),
    ('ContinentNumber', ['contID']),
    ('ResidualMomentum', ['rs1', 'rs2', 'rs3']),
    ('ResidualContinuity', ['rsc']),
    ('Stress', ['sII']),
    ('PrincipalStressAxis', ['sx1', 'sx2', 'sx3']),
    ('StrainRate', ['edot']),
    ('Primordial', ['prim']),
))

SFIELD = OrderedDict((
    ('topo_top', Varf('Topography at top', 'm')),
    ('topo_bot', Varf('Topography at bottom', 'm')),
    ('geoid_top', Varf('Geoid at top', 'm')),
    ('geoid_bot', Varf('Geoid at bottom', 'm')),
    ('topo_g_top', Varf('Topography for geoid at top', 'm')),
    ('topo_g_bot', Varf('Topography for geoid at bottom', 'm')),
    ('ftop', Varf('Heat flux at top', 'W/m2')),
    ('fbot', Varf('Heat flux at bottom', 'W/m2')),
    ('fstop', Varf('Heat flux from spectrum at top', 'W/m2')),
    ('fsbot', Varf('Heat flux from spectrum at bottom', 'W/m2')),
    ('crust', Varf('Crustal thickness', 'm')),
))

SFIELD_FILES = OrderedDict((
    ('cs', ['topo_bot', 'topo_top']),
    ('g', ['geoid_bot', 'geoid_top']),
    ('csg', ['topo_g_bot', 'topo_g_top']),
    ('hf', ['fbot', 'ftop']),
    ('hfs', ['fsbot', 'fstop']),
    ('cr', ['crust']),
))

SFIELD_FILES_H5 = OrderedDict((
    ('BottomTopography', ['topo_bot']),
    ('SurfaceTopography', ['topo_top']),
    ('BottomGeoid', ['geoid_bot']),
    ('TopGeoid', ['geoid_top']),
    ('BottomCSGeoid', ['topo_g_bot']),
    ('TopCSGeoid', ['topo_g_top']),
    ('BottomHeatFlux', ['fbot']),
    ('TopHeatFlux', ['ftop']),
    ('BottomHFSpectrum', ['fsbot']),
    ('TopHFSpectrum', ['fstop']),
    ('CrustThickness', ['crust']),
))

Varr = namedtuple('Varr', ['description', 'kind', 'dim'])
RPROF = OrderedDict((
    ('r', Varr('Radial coordinate', 'Radius', 'm')),
    ('Tmean', Varr('Temperature', 'Temperature', 'K')),
    ('Tmin', Varr('Min temperature', 'Temperature', 'K')),
    ('Tmax', Varr('Max temperature', 'Temperature', 'K')),
    ('vrms', Varr('rms velocity', 'Velocity', 'm/s')),
    ('vmin', Varr('Min velocity', 'Velocity', 'm/s')),
    ('vmax', Varr('Max velocity', 'Velocity', 'm/s')),
    ('vzabs', Varr('Radial velocity', 'Velocity', 'm/s')),
    ('vzmin', Varr('Min radial velocity', 'Velocity', 'm/s')),
    ('vzmax', Varr('Max radial velocity', 'Velocity', 'm/s')),
    ('vhrms', Varr('Horizontal velocity', 'Velocity', 'm/s')),
    ('vhmin', Varr('Min horiz velocity', 'Velocity', 'm/s')),
    ('vhmax', Varr('Max horiz velocity', 'Velocity', 'm/s')),
    ('etalog', Varr('Viscosity', 'Viscosity', 'Pa')),
    ('etamin', Varr('Min viscosity', 'Viscosity', 'Pa')),
    ('etamax', Varr('Max viscosity', 'Viscosity', 'Pa')),
    ('elog', Varr('Strain rate', 'Strain rate', '1/s')),
    ('emin', Varr('Min strain rate', 'Strain rate', '1/s')),
    ('emax', Varr('Max strain rate', 'Strain rate', '1/s')),
    ('slog', Varr('Stress', 'Stress', 'Pa')),
    ('smin', Varr('Min stress', 'Stress', 'Pa')),
    ('smax', Varr('Max stress', 'Stress', 'Pa')),
    ('whrms', Varr('Horizontal vorticity', 'Vorticity', '1/s')),
    ('whmin', Varr('Min horiz vorticity', 'Vorticity', '1/s')),
    ('whmax', Varr('Max horiz vorticity', 'Vorticity', '1/s')),
    ('wzrms', Varr('Radial vorticity', 'Vorticity', '1/s')),
    ('wzmin', Varr('Min radial vorticity', 'Vorticity', '1/s')),
    ('wzmax', Varr('Max radial vorticity', 'Vorticity', '1/s')),
    ('drms', Varr('Divergence', 'Divergence', '1/s')),
    ('dmin', Varr('Min divergence', 'Divergence', '1/s')),
    ('dmax', Varr('Max divergence', 'Divergence', '1/s')),
    ('enadv', Varr('Advection', 'Heat flux', 'W/m2')),
    ('endiff', Varr('Diffusion', 'Heat flux', 'W/m2')),
    ('enradh', Varr('Radiogenic heating', 'Heat flux', 'W/m2')),
    ('enviscdiss', Varr('Viscous dissipation', 'Heat flux', 'W/m2')),
    ('enadiabh', Varr('Adiabatic heating', 'Heat flux', 'W/m2')),
    ('cmean', Varr('Concentration', 'Concentration', '1')),
    ('cmin', Varr('Min concentration', 'Concentration', '1')),
    ('cmax', Varr('Max concentration', 'Concentration', '1')),
    ('rhomean', Varr('Density', 'Density', 'kg/m3')),
    ('rhomin', Varr('Min density', 'Density', 'kg/m3')),
    ('rhomax', Varr('Max density', 'Density', 'kg/m3')),
    ('airmean', Varr('Air', 'Air', '1')),
    ('airmin', Varr('Min air', 'Air', '1')),
    ('airmax', Varr('Max air', 'Air', '1')),
    ('primmean', Varr('Primordial', 'Concentration', '1')),
    ('primmin', Varr('Min primordial', 'Concentration', '1')),
    ('primmax', Varr('Max primordial', 'Concentration', '1')),
    ('ccmean', Varr('Continental crust', 'Crust', '1')),
    ('ccmin', Varr('Min continental crust', 'Crust', '1')),
    ('ccmax', Varr('Max continental crust', 'Crust', '1')),
    ('fmeltmean', Varr('Melt fraction', 'Melt fraction', '1')),
    ('fmeltmin', Varr('Min melt fraction', 'Melt fraction', '1')),
    ('fmeltmax', Varr('Max melt fraction', 'Melt fraction', '1')),
    ('metalmean', Varr('Metal', 'Concentration', '1')),
    ('metalmin', Varr('Min metal', 'Concentration', '1')),
    ('metalmax', Varr('Max metal', 'Concentration', '1')),
    ('gsmean', Varr('Grain size', 'Grain size', 'm')),
    ('gsmin', Varr('Min grain size', 'Grain size', 'm')),
    ('gsmax', Varr('Max grain', 'Grain size', 'm')),
    ('viscdisslog', Varr('Viscous dissipation', 'Heat flux', 'W/m2')),
    ('viscdissmin', Varr('Min visc dissipation', 'Heat flux', 'W/m2')),
    ('viscdissmax', Varr('Max visc dissipation', 'Heat flux', 'W/m2')),
    ('advtot', Varr('Advection', 'Heat flux', 'W/m2')),
    ('advdesc', Varr('Downward advection', 'Heat flux', 'W/m2')),
    ('advasc', Varr('Upward advection', 'Heat flux', 'W/m2')),
))

RPROF_EXTRA = OrderedDict((
    ('dr', Varr(processing.delta_r, 'dr', 'm')),
    ('diff', Varr(processing.diff_prof, 'Heat flux', 'W/m2')),
    ('diffs', Varr(processing.diffs_prof, 'Heat flux', 'W/m2')),
    ('advts', Varr(processing.advts_prof, 'Heat flux', 'W/m2')),
    ('advds', Varr(processing.advds_prof, 'Heat flux', 'W/m2')),
    ('advas', Varr(processing.advas_prof, 'Heat flux', 'W/m2')),
    ('energy', Varr(processing.energy_prof, 'Heat flux', 'W/m2')),
    ('ciover', Varr(processing.init_c_overturn, 'Concentration', '1')),
    ('cfover', Varr(processing.c_overturned, 'Concentration', '1')),
    ('advth', Varr(processing.advth, 'Heat Flux', 'W/m2')),
))

Vart = namedtuple('Vart', ['description', 'kind', 'dim'])
TIME = OrderedDict((
    ('t', Vart('Time', 'Time', 's')),
    ('ftop', Vart('Heat flux at top', 'Heat flux', 'W/m2')),
    ('fbot', Vart('Heat flux at bottom', 'Heat flux', 'W/m2')),
    ('Tmin', Vart('Min temperature', 'Temperature', 'K')),
    ('Tmean', Vart('Temperature', 'Temperature', 'K')),
    ('Tmax', Vart('Max temperature', 'Temperature', 'K')),
    ('vmin', Vart('Min velocity', 'Velocity', 'm/s')),
    ('vrms', Vart('rms velocity', 'Velocity', 'm/s')),
    ('vmax', Vart('Max velocity', 'Velocity', 'm/s')),
    ('etamin', Vart('Min viscosity', 'Viscosity', 'Pa.s')),
    ('etamean', Vart('Viscosity', 'Viscosity', 'Pa.s')),
    ('etamax', Vart('Max viscosity', 'Viscosity', 'Pa.s')),
    ('Raeff', Vart('Effective Ra', r'$\mathrm{Ra}$', '1')),
    ('Nutop', Vart('Nusselt at top', r'$\mathrm{Nu}$', '1')),
    ('Nubot', Vart('Nusselt at bot', r'$\mathrm{Nu}$', '1')),
    ('Cmin', Vart('Min concentration', 'Concentration', '1')),
    ('Cmean', Vart('Concentration', 'Concentration', '1')),
    ('Cmax', Vart('Max concentration', 'Concentration', '1')),
    ('moltenf_mean', Vart('Molten fraction', 'Fraction', '1')),
    ('moltenf_max', Vart('Max molten fraction', 'Fraction', '1')),
    ('erupt_rate', Vart('Eruption rate', 'Eruption rate', '1/s')),
    ('erupt_tot', Vart('Erupta total', 'Eruption rate', '1/s')),
    ('erupt_heat', Vart('Erupta heat', 'Eruption rate', '1/s')),
    ('entrainment', Vart('Entrainment', 'Eruption rate', '1/s')),
    ('Cmass_error', Vart('Error on Cmass', 'Error', '1')),
    ('H_int', Vart('Internal heating', 'Internal heating', 'W/m3')),
    ('r_ic', Vart('Inner core radius', 'Inner core radius', 'm')),
    ('topT_val', Vart('Temperature at top', 'Temperature', 'K')),
    ('botT_val', Vart('Temperature at bottom', 'Temperature', 'K')),
))

TIME_EXTRA = OrderedDict((
    ('dt', Vart(processing.dtime, 'dt', 's')),
    ('dTdt', Vart(processing.dt_dt, r'dT/dt', 'K/s')),
    ('ebalance', Vart(processing.ebalance, r'$\mathrm{Nu}$', '1')),
    ('mobility', Vart(processing.mobility, 'Mobility', '1')),
))

REFSTATE = OrderedDict((
    ('z', Varr('z position', 'z position', 'm')),
    ('T', Varr('Temperature', 'Temperature', 'K')),
    ('rho', Varr('Density', 'Density', 'kg/m3')),
    ('expan', Varr('Thermal expansivity', 'Thermal expansivity', '1/K')),
    ('Cp', Varr('Heat capacity', 'Heat capacity', 'J/(kg.K)')),
    ('Tcond', Varr('Conductivity', 'Conductivity', 'W/(m.K)')),
    ('P', Varr('Pressure', 'Pressure', 'Pa')),
    ('grav', Varr('Gravity', 'Gravity', 'm/s2')),
))

Varp = namedtuple('Varp', ['description'])
PLATES = OrderedDict((
    ('c', Varp('Composition')),
    ('eta', Varp('Viscosity')),
    ('sc', Varp('Topography')),
    ('age', Varp('Age')),
    ('str', Varp('Stress')),
    ('sx', Varp('Principal deviatoric stress')),
    ('ed', Varp('Strain rate')),
))

SCALES = {
    'm': attrgetter('length'),
    'kg/m3': attrgetter('density'),
    'K': attrgetter('temperature'),
    'W/m2': attrgetter('heat_flux'),
    'Pa': attrgetter('stress'),
    'Pa.s': attrgetter('dyn_visc'),
    's': attrgetter('time'),
    'W/(m.K)': attrgetter('th_cond'),
    'm2/s': attrgetter('th_diff'),
    'W/m3': lambda scl: scl.power / scl.length**3,
    '1/s': lambda scl: 1 / scl.time,
    'K/s': lambda scl: scl.temperature / scl.time,
    'm/s': lambda scl: scl.length / scl.time,
    'm/s2': lambda scl: scl.length / scl.time**2,
}

PREFIXES = ('k', 'M', 'G')



"""Plot radial profiles."""
import matplotlib.pyplot as plt

from . import conf, misc
from .stagyydata import StagyyData


def plot_rprofs(rprofs, lovs):
    """Plot requested radial profiles.

    Args:
        rprofs (:class:`~stagpy._step._Rprofs`): a radial profile viewer
            such as :attr:`Step.rprofs` and :attr:`_StepsView.rprofs_averaged`.
        lovs (nested list of str): nested list of profile names such as the one
            produced by :func:`stagpy.misc.list_of_vars`.
    """
    try:
        stepstr = rprofs.steps.stepstr
    except AttributeError:
        stepstr = str(rprofs.step.istep)
    sdat = rprofs.step.sdat

    for vfig in lovs:
        fig, axes = plt.subplots(ncols=len(vfig), sharey=True,
                                 figsize=(4 * len(vfig), 6))
        axes = [axes] if len(vfig) == 1 else axes
        fname = 'rprof_'
        for iplt, vplt in enumerate(vfig):
            xlabel = None
            profs_on_plt = (rprofs[rvar] for rvar in vplt)
            fname += '_'.join(vplt) + '_'
            for ivar, (rprof, rad, meta) in enumerate(profs_on_plt):
                if conf.rprof.depth:
                    rad = sdat.scale(rprofs.bounds[1], 'm')[0] - rad
                axes[iplt].plot(rprof, rad,
                                conf.rprof.style,
                                label=meta.description)
                if xlabel is None:
                    xlabel = meta.kind
                elif xlabel != meta.kind:
                    xlabel = ''
            if ivar == 0:
                xlabel = meta.description
            if xlabel:
                _, unit = sdat.scale(1, meta.dim)
                xlabel += f' ({unit})' if unit else ''
                axes[iplt].set_xlabel(xlabel)
            if vplt[0][:3] == 'eta':  # list of log variables
                axes[iplt].set_xscale('log')
            axes[iplt].set_xlim(left=conf.plot.vmin, right=conf.plot.vmax)
            if ivar:
                axes[iplt].legend()
        if conf.rprof.depth:
            axes[0].invert_yaxis()
        ylabel = 'Depth' if conf.rprof.depth else 'Radius'
        _, unit = sdat.scale(1, 'm')
        ylabel += f' ({unit})' if unit else ''
        axes[0].set_ylabel(ylabel)
        misc.saveplot(fig, fname + stepstr)


def plot_grid(step):
    """Plot cell position and thickness.

    The figure is call grid_N.pdf where N is replace by the step index.

    Args:
        step (:class:`~stagpy.stagyydata._Step`): a step of a StagyyData
            instance.
    """
    drad, rad, _ = step.rprofs['dr']
    _, unit = step.sdat.scale(1, 'm')
    if unit:
        unit = f' ({unit})'
    fig, (ax1, ax2) = plt.subplots(2, sharex=True)
    ax1.plot(rad, '-ko')
    ax1.set_ylabel('$r$' + unit)
    ax2.plot(drad, '-ko')
    ax2.set_ylabel('$dr$' + unit)
    ax2.set_xlim([-0.5, len(rad) - 0.5])
    ax2.set_xlabel('Cell number')
    misc.saveplot(fig, 'grid', step.istep)


def cmd():
    """Implementation of rprof subcommand.

    Other Parameters:
        conf.rprof
        conf.core
    """
    sdat = StagyyData()

    if conf.rprof.grid:
        for step in sdat.walk.filter(rprofs=True):
            plot_grid(step)

    lovs = misc.list_of_vars(conf.rprof.plot)
    if not lovs:
        return

    if conf.rprof.average:
        plot_rprofs(sdat.walk.rprofs_averaged, lovs)
    else:
        for step in sdat.walk.filter(rprofs=True):
            plot_rprofs(step.rprofs, lovs)



import pytest
import f90nml
import pandas
import stagpy.stagyydata
import stagpy._step
import stagpy.error


def test_sdat_path(example_dir, sdat):
    assert sdat.path == example_dir


def test_sdat_deflt_nfields_max(sdat):
    assert sdat.nfields_max == 50


def test_sdat_set_nfields_max(sdat):
    sdat.nfields_max = 6
    assert sdat.nfields_max == 6


def test_sdat_set_nfields_max_none(sdat):
    sdat.nfields_max = None
    assert sdat.nfields_max is None


def test_sdat_set_nfields_max_invalid(sdat):
    with pytest.raises(stagpy.error.InvalidNfieldsError) as err:
        sdat.nfields_max = 5
    assert err.value.nfields == 5


def test_sdat_par(sdat):
    assert isinstance(sdat.par, f90nml.namelist.Namelist)


def test_sdat_tseries(sdat):
    assert sdat.tseries['Tmean'].time is sdat.tseries.time


def test_sdat_rtimes(sdat):
    assert isinstance(sdat.rtimes, pandas.DataFrame)


def test_sdat_walk_dflt(sdat):
    wlk = iter(sdat.walk)
    assert next(wlk) is sdat.snaps[-1]
    with pytest.raises(StopIteration):
        next(wlk)


def test_steps_iter(sdat):
    assert sdat.steps[:] == sdat.steps


def test_snaps_iter(sdat):
    assert sdat.snaps[:] == sdat.snaps


def test_filter_snap(sdat):
    snaps = (s for s in sdat.snaps if s)
    assert sdat.steps.filter(snap=True) == snaps


def test_filter_even_steps(sdat):
    even = sdat.steps.filter(func=lambda step: step.istep % 2 == 0)
    assert sdat.steps[::2] == even


def test_snaps_last(sdat):
    assert sdat.snaps[-1] is sdat.snaps[len(sdat.snaps) - 1]


def test_snaps_empty(sdat):
    with pytest.raises(stagpy.error.InvalidSnapshotError):
        sdat.snaps[len(sdat.snaps)]


def test_step_is_snap(sdat):
    istep = sdat.snaps[-1].istep
    assert sdat.steps[istep] is sdat.snaps[-1]


def test_step_sdat(sdat):
    assert all(s.sdat is sdat for s in sdat.steps)


def test_step_istep(sdat):
    assert all(s is sdat.steps[s.istep] for s in sdat.steps)


def test_geom_refs(step):
    assert step.geom is step.fields.geom


def test_geom(step):
    assert step.geom.twod
    assert not step.geom.threed
    assert not step.geom.yinyang
    assert step.geom.cartesian is not step.geom.spherical



"""Define configuration variables for StagPy.

See :mod:`stagpy.args` for additional definitions related to the command line
interface.
"""

from collections import OrderedDict
import pathlib

from loam.manager import ConfOpt as Conf
from loam.tools import switch_opt, config_conf_section, set_conf_opt


def _actual_index(arg):
    """Turn a string in a integer or slice."""
    if ':' in arg:
        idxs = arg.split(':')
        if len(idxs) > 3:
            raise ValueError(f'{arg} is an invalid slice')
        idxs[0] = int(idxs[0]) if idxs[0] else None
        idxs[1] = int(idxs[1]) if idxs[1] else None
        if len(idxs) == 3:
            idxs[2] = int(idxs[2]) if idxs[2] else None
        else:
            idxs = idxs[0:2] + [1]
        return slice(*idxs)
    return int(arg)


def _index_collection(arg):
    """Build an index collection from a command line input."""
    return [_actual_index(item) for item in arg.split(',') if item]


HOME_DIR = pathlib.Path.home()
CONFIG_DIR = HOME_DIR / '.config' / 'stagpy'
CONFIG_FILE = CONFIG_DIR / 'config.toml'
CONFIG_LOCAL = pathlib.Path('.stagpy.toml')

CONF_DEF = OrderedDict()

CONF_DEF['common'] = OrderedDict((
    ('config', Conf(None, True, None, {'action': 'store_true'},
                    False, 'print config options')),
    ('set', set_conf_opt()),
))

CONF_DEF['core'] = OrderedDict((
    ('path', Conf('./', True, 'p', {},
                  True, 'path of StagYY run directory or par file', '_files')),
    ('outname', Conf('stagpy', True, 'n', {},
                     True, 'StagPy output file name prefix')),
    ('shortname', switch_opt(False, None,
                             'StagPy output file name is only prefix')),
    ('timesteps', Conf(None, True, 't',
                       {'nargs': '?', 'const': '', 'type': _index_collection},
                       False, 'timesteps slice')),
    ('snapshots', Conf(None, True, 's',
                       {'nargs': '?', 'const': '', 'type': _index_collection},
                       False, 'snapshots slice')),
))

CONF_DEF['plot'] = OrderedDict((
    ('ratio', Conf(None, True, None,
                   {'nargs': '?', 'const': 0.6, 'type': float},
                   False, 'force aspect ratio of field plot')),
    ('raster', switch_opt(True, None, 'rasterize field plots')),
    ('format', Conf('pdf', True, None, {},
                    True, 'figure format (pdf, eps, svg, png)')),
    ('vmin', Conf(None, True, None, {'type': float},
                  False, 'minimal value on plot')),
    ('vmax', Conf(None, True, None, {'type': float},
                  False, 'maximal value on plot')),
    ('cminmax', switch_opt(False, 'C', 'constant min max across plots')),
    ('mplstyle', Conf('stagpy-paper', True, None,
                      {'nargs': '?', 'const': '', 'type': str},
                      True, 'matplotlib style')),
    ('xkcd', Conf(False, False, None, {},
                  True, 'use the xkcd style')),
))

CONF_DEF['scaling'] = OrderedDict((
    ('yearins', Conf(3.154e7, False, None, {},
                     True, 'year in seconds')),
    ('ttransit', Conf(1.78e15, False, None, {},
                      True, 'transit time in My')),
    ('dimensional', switch_opt(False, None, 'use dimensional units')),
    ('time_in_y', switch_opt(True, None, 'dimensionful time is in year')),
    ('vel_in_cmpy', switch_opt(True, None,
                               'dimensionful velocity is in cm/year')),
    ('factors', Conf({'s': 'M',
                      'm': 'k',
                      'Pa': 'G'},
                     False, None, {}, True, 'custom factors')),
))

CONF_DEF['field'] = OrderedDict((
    ('plot',
        Conf('T,stream', True, 'o',
             {'nargs': '?', 'const': '', 'type': str},
             True, 'variables to plot (see stagpy var)')),
    ('perturbation', switch_opt(False, None,
                                'plot departure from average profile')),
    ('shift', Conf(None, True, None, {'type': int},
                   False, 'shift plot horizontally')),
    ('timelabel', switch_opt(False, None, 'add label with time')),
    ('interpolate', switch_opt(False, None, 'apply Gouraud shading')),
    ('colorbar', switch_opt(True, None, 'add color bar to plot')),
    ('ix', Conf(None, True, None, {'type': int},
                False, 'x-index of slice for 3D fields')),
    ('iy', Conf(None, True, None, {'type': int},
                False, 'y-index of slice for 3D fields')),
    ('iz', Conf(None, True, None, {'type': int},
                False, 'z-index of slice for 3D fields')),
    ('isocolors', Conf('', True, None, {}, True,
                       'comma-separated list of colors for isolines')),
    ('cmap',
        Conf({'T': 'RdBu_r',
              'eta': 'viridis_r',
              'rho': 'RdBu',
              'sII': 'plasma_r',
              'edot': 'Reds'},
             False, None, {}, True, 'custom colormaps')),
))

CONF_DEF['rprof'] = OrderedDict((
    ('plot',
        Conf('Tmean', True, 'o',
             {'nargs': '?', 'const': ''},
             True, 'variables to plot (see stagpy var)')),
    ('style',
        Conf('-', True, None, {},
             True, 'matplotlib line style')),
    ('average', switch_opt(False, 'a', 'plot temporal average')),
    ('grid', switch_opt(False, 'g', 'plot grid')),
    ('depth', switch_opt(False, 'd', 'depth as vertical axis')),
))

CONF_DEF['time'] = OrderedDict((
    ('plot',
        Conf('Nutop,ebalance,Nubot.Tmean', True, 'o',
             {'nargs': '?', 'const': ''},
             True, 'variables to plot (see stagpy var)')),
    ('style',
        Conf('-', True, None, {},
             True, 'matplotlib line style')),
    ('compstat',
        Conf('', True, None, {'nargs': '?', 'const': ''},
             False, 'compute mean and rms of listed variables')),
    ('tstart',
        Conf(None, True, None, {'type': float},
             False, 'beginning time')),
    ('tend',
        Conf(None, True, None, {'type': float},
             False, 'end time')),
    ('fraction',
        Conf(None, True, None, {'type': float},
             False, 'ending fraction of series to process')),
    ('marktimes',
        Conf('', True, 'M', {},
             False, 'list of times where to put a mark')),
    ('marksteps',
        Conf('', True, 'T', {'type': _index_collection},
             False, 'list of steps where to put a mark')),
    ('marksnaps',
        Conf('', True, 'S', {'type': _index_collection},
             False, 'list of snaps where to put a mark')),
))

CONF_DEF['refstate'] = OrderedDict((
    ('plot',
        Conf('T', True, 'o',
             {'nargs': '?', 'const': ''},
             True, 'variables to plot (see stagpy var)')),
    ('style',
        Conf('-', True, None, {},
             True, 'matplotlib line style')),
))

CONF_DEF['plates'] = OrderedDict((
    ('plot',
        Conf('c,eta,sc', True, 'o',
             {'nargs': '?', 'const': '', 'type': str},
             True, 'variables to plot (see stagpy var)')),
    ('vzcheck', switch_opt(False, None,
                           'activate Colin\'s version with vz checking')),
    ('timeprofile', switch_opt(False, None,
                               'nb of plates as function of time')),
    ('zoom',
        Conf(None, True, None, {'type': float},
             False, 'zoom around surface')),
    ('topomin', Conf(-40, False, None, {},
                     True, 'min topography in plots')),
    ('topomax', Conf(100, False, None, {},
                     True, 'max topography in plots')),
    ('agemin', Conf(-50, False, None, {},
                    True, 'min age in plots')),
    ('agemax', Conf(500, False, None, {},
                    True, 'max age in plots')),
    ('vmin', Conf(-5000, False, None, {},
                  True, 'min velocity in plots')),
    ('vmax', Conf(5000, False, None, {},
                  True, 'max velocity in plots')),
    ('dvmin', Conf(-250000, False, None, {},
                   True, 'min velocity derivative in plots')),
    ('dvmax', Conf(150000, False, None, {},
                   True, 'max velocity derivative in plots')),
    ('stressmin', Conf(0, False, None, {},
                       True, 'min stress in plots')),
    ('stressmax', Conf(800, False, None, {},
                       True, 'max stress in plots')),
    ('lstressmax', Conf(50, False, None, {},
                        True, 'max lithospheric stress in plots')),
))

CONF_DEF['info'] = OrderedDict((
    ('output', Conf('t,Tmean,vrms,Nutop,Nubot', True, 'o', {},
                    True, 'time series to print')),
))

CONF_DEF['var'] = OrderedDict((
    ('field', Conf(None, True, None, {'action': 'store_true'},
                   False, 'print field variables')),
    ('sfield', Conf(None, True, None, {'action': 'store_true'},
                    False, 'print surface field variables')),
    ('rprof', Conf(None, True, None, {'action': 'store_true'},
                   False, 'print rprof variables')),
    ('time', Conf(None, True, None, {'action': 'store_true'},
                  False, 'print time variables')),
    ('refstate', Conf(None, True, None, {'action': 'store_true'},
                      False, 'print refstate variables')),
    ('plates', Conf(None, True, None, {'action': 'store_true'},
                    False, 'print plates variables')),
))

CONF_DEF['config'] = config_conf_section()



"""Parse command line arguments and update :attr:`stagpy.conf`."""

from collections import OrderedDict
from inspect import isfunction

from loam.tools import set_conf_str, create_complete_files
from loam.cli import Subcmd, CLIManager

from . import __doc__ as doc_module
from . import conf, PARSING_OUT, load_mplstyle
from . import commands, field, rprof, time_series, refstate, plates
from .misc import baredoc
from .config import CONFIG_DIR


def _sub(cmd, *sections):
    """Build Subcmd instance."""
    cmd_func = cmd if isfunction(cmd) else cmd.cmd
    return Subcmd(baredoc(cmd), *sections, func=cmd_func)


def _bare_cmd():
    """Print help message when no arguments are given."""
    print(doc_module)
    print('Run `stagpy -h` for usage')


SUB_CMDS = OrderedDict((
    ('common_', Subcmd(doc_module, 'common', func=_bare_cmd)),
    ('field', _sub(field, 'core', 'plot', 'scaling')),
    ('rprof', _sub(rprof, 'core', 'plot', 'scaling')),
    ('time', _sub(time_series, 'core', 'plot', 'scaling')),
    ('refstate', _sub(refstate, 'core', 'plot')),
    ('plates', _sub(plates, 'core', 'plot', 'scaling')),
    ('info', _sub(commands.info_cmd, 'core', 'scaling')),
    ('var', _sub(commands.var_cmd)),
    ('version', _sub(commands.version_cmd)),
    ('config', _sub(commands.config_cmd)),
))


def parse_args(arglist=None):
    """Parse cmd line arguments.

    Update :attr:`stagpy.conf` accordingly.

    Args:
        arglist (list of str): the list of cmd line arguments. If set to
            None, the arguments are taken from :attr:`sys.argv`.

    Returns:
        function: the function implementing the sub command to be executed.
    """
    climan = CLIManager(conf, **SUB_CMDS)

    create_complete_files(climan, CONFIG_DIR, 'stagpy', zsh_sourceable=True)

    cmd_args, all_subs = climan.parse_args(arglist)
    sub_cmd = cmd_args.loam_sub_name

    if sub_cmd is None:
        return cmd_args.func

    if sub_cmd != 'config':
        commands.report_parsing_problems(PARSING_OUT)

    if conf.common.set:
        set_conf_str(conf, conf.common.set)

    if conf.common.config:
        commands.config_pp(all_subs)

    load_mplstyle()

    return cmd_args.func



from stagpy import processing


def test_dt_dt(sdat):
    dtdt, time = processing.dt_dt(sdat)
    assert dtdt.shape == time.shape == (sdat.tseries.time.shape[0] - 1,)


def test_ebalance(sdat):
    ebal, time = processing.ebalance(sdat)
    assert ebal.shape == time.shape == (sdat.tseries.time.shape[0] - 1,)


def test_r_edges(step):
    assert step.rprofs.walls.shape == (step.geom.nztot + 1,)


def test_delta_r(step):
    thick, _ = processing.delta_r(step)
    assert thick.shape == (step.geom.nztot,)


def test_diff_prof(step):
    diff, rpos = processing.diff_prof(step)
    assert diff.shape == rpos.shape == (step.geom.nztot + 1,)


def test_diffs_prof(step):
    diff, rpos = processing.diffs_prof(step)
    assert diff.shape == rpos.shape == (step.geom.nztot + 1,)


def test_advts_prof(step):
    adv, _ = processing.advts_prof(step)
    assert adv.shape == (step.geom.nztot,)


def test_advds_prof(step):
    adv, _ = processing.advds_prof(step)
    assert adv.shape == (step.geom.nztot,)


def test_advas_prof(step):
    adv, _ = processing.advas_prof(step)
    assert adv.shape == (step.geom.nztot,)


def test_energy_prof(step):
    eng, rpos = processing.energy_prof(step)
    assert eng.shape == rpos.shape == (step.geom.nztot + 1,)


def test_stream_function(step):
    psi = processing.stream_function(step)
    assert psi.shape[1:3] == step.fields['v3'].shape[1:3]



"""Plots time series."""

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from . import conf, misc
from .error import InvalidTimeFractionError
from .stagyydata import StagyyData


def _collect_marks(sdat):
    """Concatenate mark* config variable."""
    times = set(conf.time.marktimes.replace(',', ' ').split())
    times = list(map(float, times))
    times.extend(step.timeinfo['t']
                 for step in sdat.snaps[conf.time.marksnaps])
    times.extend(step.timeinfo['t']
                 for step in sdat.steps[conf.time.marksteps])
    return times


def plot_time_series(sdat, lovs):
    """Plot requested time series.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
        lovs (nested list of str): nested list of series names such as
            the one produced by :func:`stagpy.misc.list_of_vars`.

    Other Parameters:
        conf.time.tstart: the starting time.
        conf.time.tend: the ending time.
    """
    time_marks = _collect_marks(sdat)
    for vfig in lovs:
        tstart = conf.time.tstart
        tend = conf.time.tend
        fig, axes = plt.subplots(nrows=len(vfig), sharex=True,
                                 figsize=(12, 2 * len(vfig)))
        axes = [axes] if len(vfig) == 1 else axes
        fname = ['time']
        for iplt, vplt in enumerate(vfig):
            ylabel = None
            series_on_plt = [
                sdat.tseries.tslice(tvar, conf.time.tstart, conf.time.tend)
                for tvar in vplt]
            ptstart = min(time[0] for _, time, _ in series_on_plt)
            ptend = max(time[-1] for _, time, _ in series_on_plt)
            tstart = ptstart if tstart is None else min(ptstart, tstart)
            tend = ptend if tend is None else max(ptend, tend)
            fname.extend(vplt)
            for ivar, (series, time, meta) in enumerate(series_on_plt):
                axes[iplt].plot(time, series, conf.time.style,
                                label=meta.description)
                lbl = meta.kind
                if ylabel is None:
                    ylabel = lbl
                elif ylabel != lbl:
                    ylabel = ''
            if ivar == 0:
                ylabel = meta.description
            if ylabel:
                _, unit = sdat.scale(1, meta.dim)
                if unit:
                    ylabel += f' ({unit})'
                axes[iplt].set_ylabel(ylabel)
            if vplt[0][:3] == 'eta':  # list of log variables
                axes[iplt].set_yscale('log')
            axes[iplt].set_ylim(bottom=conf.plot.vmin, top=conf.plot.vmax)
            if ivar:
                axes[iplt].legend()
            axes[iplt].tick_params()
            for time_mark in time_marks:
                axes[iplt].axvline(time_mark, color='black', linestyle='--')
        _, unit = sdat.scale(1, 's')
        if unit:
            unit = f' ({unit})'
        axes[-1].set_xlabel('Time' + unit)
        axes[-1].set_xlim(tstart, tend)
        axes[-1].tick_params()
        misc.saveplot(fig, '_'.join(fname))


def compstat(sdat, *names, tstart=None, tend=None):
    """Compute statistics from series output by StagYY.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
        names (str): variables whose statistics should be computed.
        tstart (float): starting time. Set to None to start at the beginning of
            available data.
        tend (float): ending time. Set to None to stop at the end of available
            data.
    Returns:
        :class:`pandas.DataFrame`: computed statistics with 'mean' and 'rms' as
            index and variable names as columns.
    """
    stats = pd.DataFrame(columns=names, index=['mean', 'rms'])
    for name in names:
        data, time, _ = sdat.tseries.tslice(name, tstart, tend)
        delta_time = time[-1] - time[0]
        mean = np.trapz(data, x=time) / delta_time
        stats.loc['mean', name] = mean
        stats.loc['rms', name] = np.sqrt(np.trapz((data - mean)**2, x=time) /
                                         delta_time)
    return stats


def cmd():
    """Implementation of time subcommand.

    Other Parameters:
        conf.time
        conf.core
    """
    sdat = StagyyData()
    if sdat.tseries is None:
        return

    if conf.time.fraction is not None:
        if not 0 < conf.time.fraction <= 1:
            raise InvalidTimeFractionError(conf.time.fraction)
        conf.time.tend = None
        t_0 = sdat.tseries.time[0]
        t_f = sdat.tseries.time[-1]
        conf.time.tstart = (t_0 * conf.time.fraction +
                            t_f * (1 - conf.time.fraction))

    lovs = misc.list_of_vars(conf.time.plot)
    if lovs:
        plot_time_series(sdat, lovs)

    if conf.time.compstat:
        names = conf.time.compstat.replace(',', ' ').split()
        stats = compstat(sdat, *names, tstart=conf.time.tstart,
                         tend=conf.time.tend)
        stats.to_csv('statistics.dat')



"""Computations of phyvars that are not output by StagYY.

Time series are returned along with the time at which the variables are
evaluated. Radial profiles are returned along with the radial positions at
which the variables are evaluated.
"""

import numpy as np
from scipy import integrate

from .error import NotAvailableError


def dtime(sdat):
    """Time increment dt.

    Compute dt as a function of time.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: dt and time arrays.
    """
    time = sdat.tseries.time
    return time[1:] - time[:-1], time[:-1]


def dt_dt(sdat):
    """Derivative of temperature.

    Compute dT/dt as a function of time using an explicit Euler scheme.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: derivative of temperature and time
        arrays.
    """
    temp, time, _ = sdat.tseries['Tmean']
    dtdt = (temp[1:] - temp[:-1]) / (time[1:] - time[:-1])
    return dtdt, time[:-1]


def ebalance(sdat):
    """Energy balance.

    Compute Nu_t - Nu_b + V*dT/dt as a function of time using an explicit
    Euler scheme. This should be zero if energy is conserved.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: energy balance and time arrays.
    """
    rbot, rtop = sdat.steps[-1].rprofs.bounds
    if rbot != 0:  # spherical
        coefsurf = (rtop / rbot)**2
        volume = rbot * ((rtop / rbot)**3 - 1) / 3
    else:
        coefsurf = 1.
        volume = 1.
    dtdt, time = dt_dt(sdat)
    ftop = sdat.tseries['ftop'].values * coefsurf
    fbot = sdat.tseries['fbot'].values
    radio = sdat.tseries['H_int'].values
    ebal = ftop[1:] - fbot[1:] + volume * (dtdt - radio[1:])
    return ebal, time


def mobility(sdat):
    """Plates mobility.

    Compute the ratio vsurf / vrms.

    Args:
        sdat (:class:`~stagpy.stagyydata.StagyyData`): a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: mobility and time arrays.
    """
    time = []
    mob = []
    for step in sdat.steps.filter(rprofs=True):
        time.append(step.timeinfo['t'])
        mob.append(step.rprofs['vrms'].values[-1] / step.timeinfo['vrms'])
    return np.array(mob), np.array(time)


def delta_r(step):
    """Cells thickness.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the thickness of the cells and radius.
    """
    edges = step.rprofs.walls
    return (edges[1:] - edges[:-1]), step.rprofs.centers


def _scale_prof(step, rprof, rad=None):
    """Scale profile to take sphericity into account."""
    rbot, rtop = step.rprofs.bounds
    if rbot == 0:  # not spherical
        return rprof
    if rad is None:
        rad = step.rprofs.centers
    return rprof * (2 * rad / (rtop + rbot))**2


def diff_prof(step):
    """Diffusion.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the diffusion and radius.
    """
    rbot, rtop = step.rprofs.bounds
    rad = step.rprofs.centers
    tprof = step.rprofs['Tmean'].values
    diff = (tprof[:-1] - tprof[1:]) / (rad[1:] - rad[:-1])
    # assume tbot = 1
    diff = np.insert(diff, 0, (1 - tprof[0]) / (rad[0] - rbot))
    # assume ttop = 0
    diff = np.append(diff, tprof[-1] / (rtop - rad[-1]))
    return diff, step.rprofs.walls


def diffs_prof(step):
    """Scaled diffusion.

    This computation takes sphericity into account if necessary.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the diffusion and radius.
    """
    diff, rad = diff_prof(step)
    return _scale_prof(step, diff, rad), rad


def advts_prof(step):
    """Scaled advection.

    This computation takes sphericity into account if necessary.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the scaled advection and radius.
    """
    return _scale_prof(step, step.rprofs['advtot'].values), step.rprofs.centers


def advds_prof(step):
    """Scaled downward advection.

    This computation takes sphericity into account if necessary.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the scaled downward advection and
            radius.
    """
    return (_scale_prof(step, step.rprofs['advdesc'].values),
            step.rprofs.centers)


def advas_prof(step):
    """Scaled upward advection.

    This computation takes sphericity into account if necessary.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the scaled upward advection and radius.
    """
    return _scale_prof(step, step.rprofs['advasc'].values), step.rprofs.centers


def energy_prof(step):
    """Energy flux.

    This computation takes sphericity into account if necessary.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the energy flux and radius.
    """
    diff, rad = diffs_prof(step)
    adv, _ = advts_prof(step)
    return (diff + np.append(adv, 0)), rad


def advth(step):
    """Theoretical advection.

    This compute the theoretical profile of total advection as function of
    radius.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the theoretical advection and radius.
    """
    rbot, rtop = step.rprofs.bounds
    rmean = 0.5 * (rbot + rtop)
    rad = step.rprofs.centers
    radio = step.timeinfo['H_int']
    if rbot != 0:  # spherical
        th_adv = -(rtop**3 - rad**3) / rmean**2 / 3
    else:
        th_adv = rad - rtop
    th_adv *= radio
    th_adv += step.timeinfo['Nutop']
    return th_adv, rad


def init_c_overturn(step):
    """Initial concentration.

    This compute the resulting composition profile if fractional
    crystallization of a SMO is assumed.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the composition and radius.
    """
    rbot, rtop = step.rprofs.bounds
    xieut = step.sdat.par['tracersin']['fe_eut']
    k_fe = step.sdat.par['tracersin']['k_fe']
    xi0l = step.sdat.par['tracersin']['fe_cont']
    xi0s = k_fe * xi0l
    xired = xi0l / xieut
    rsup = (rtop**3 - xired**(1 / (1 - k_fe)) *
            (rtop**3 - rbot**3))**(1 / 3)

    def initprof(rpos):
        """Theoretical initial profile."""
        if rpos < rsup:
            return xi0s * ((rtop**3 - rbot**3) /
                           (rtop**3 - rpos**3))**(1 - k_fe)
        return xieut

    rad = np.linspace(rbot, rtop, 500)
    initprof = np.vectorize(initprof)
    return initprof(rad), rad


def c_overturned(step):
    """Theoretical overturned concentration.

    This compute the resulting composition profile if fractional
    crystallization of a SMO is assumed and then a purely radial
    overturn happens.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        tuple of :class:`numpy.array`: the composition and radius.
    """
    rbot, rtop = step.rprofs.bounds
    cinit, rad = init_c_overturn(step)
    radf = (rtop**3 + rbot**3 - rad**3)**(1 / 3)
    return cinit, radf


def stream_function(step):
    """Stream function.

    Args:
        step (:class:`~stagpy._step.Step`): a step of a StagyyData instance.
    Returns:
        :class:`numpy.array`: the stream function field, with four dimensions:
        x-direction, y-direction, z-direction and block.
    """
    if step.geom.twod_yz:
        x_coord = step.geom.y_walls
        v_x = step.fields['v2'][0, :, :, 0]
        v_z = step.fields['v3'][0, :, :, 0]
        shape = (1, v_x.shape[0], v_x.shape[1], 1)
    elif step.geom.twod_xz and step.geom.cartesian:
        x_coord = step.geom.x_walls
        v_x = step.fields['v1'][:, 0, :, 0]
        v_z = step.fields['v3'][:, 0, :, 0]
        shape = (v_x.shape[0], 1, v_x.shape[1], 1)
    else:
        raise NotAvailableError('Stream function only implemented in '
                                '2D cartesian and spherical annulus')
    psi = np.zeros_like(v_x)
    if step.geom.spherical:  # YZ annulus
        # positions
        r_nc = step.rprofs.centers  # numerical centers
        r_pc = step.geom.r_centers  # physical centers
        r_nw = step.rprofs.walls[:2]  # numerical walls of first cell
        # vz at center of bottom cells
        vz0 = ((r_nw[1] - r_nc[0]) * v_z[:, 0] +
               (r_nc[0] - r_nw[0]) * v_z[:, 1]) / (r_nw[1] - r_nw[0])
        psi[1:, 0] = -integrate.cumtrapz(r_pc[0]**2 * vz0, x=x_coord)
        # vx at center
        vxc = (v_x + np.roll(v_x, -1, axis=0)) / 2
        for i_x in range(len(x_coord)):
            psi[i_x, 1:] = psi[i_x, 0] + \
                integrate.cumtrapz(r_pc * vxc[i_x], x=r_nc)
    else:  # assume cartesian geometry
        z_nc = step.geom.r_centers
        z_nw = step.rprofs.walls[:2]
        vz0 = ((z_nw[1] - z_nc[0]) * v_z[:, 0] +
               (z_nc[0] - z_nw[0]) * v_z[:, 1]) / (z_nw[1] - z_nw[0])
        psi[1:, 0] = -integrate.cumtrapz(vz0, x=x_coord)
        # vx at center
        vxc = (v_x + np.roll(v_x, -1, axis=0)) / 2
        for i_x in range(len(x_coord)):
            psi[i_x, 1:] = psi[i_x, 0] + \
                integrate.cumtrapz(vxc[i_x], x=z_nc)
    if step.geom.twod_xz:
        psi = - psi
    psi = np.reshape(psi, shape)
    return psi



"""StagPy is a tool to postprocess StagYY output files.

StagPy is both a CLI tool and a powerful Python library. See the
documentation at
https://stagpy.readthedocs.io/en/stable/

If the environment variable STAGPY_ISOLATED is set to a truthy value, StagPy
does not attempt to read any configuration file (including mplstyle).

When using the CLI interface, if the environment variable STAGPY_DEBUG is set
to a truthy value, warnings are issued normally and StagpyError are raised.
Otherwise, warnings are ignored and only a short form of encountered
StagpyErrors is printed.

Truthy values for environment variables are 'true', 't', 'yes', 'y', 'on', '1',
and uppercase versions of those.
"""

import importlib
import os
import pathlib
import shutil
import signal
import sys

from pkg_resources import get_distribution, DistributionNotFound
from setuptools_scm import get_version
from loam.manager import ConfigurationManager

from . import config


def _env(var):
    """Return whether var is set to True."""
    val = os.getenv(var, default='').lower()
    return val in ('true', 't', 'yes', 'y', 'on', '1')


DEBUG = _env('STAGPY_DEBUG')
ISOLATED = _env('STAGPY_ISOLATED')


def sigint_handler(*_):
    """Handler of SIGINT signal.

    It is set when you use StagPy as a command line tool to handle gracefully
    keyboard interruption.
    """
    print('\nSo long, and thanks for all the fish.')
    sys.exit()


def _check_config():
    """Create config files as necessary."""
    config.CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    verfile = config.CONFIG_DIR / '.version'
    uptodate = verfile.is_file() and verfile.read_text() == __version__
    if not uptodate:
        verfile.write_text(__version__)
    if not (uptodate and config.CONFIG_FILE.is_file()):
        conf.create_config_(update=True)
    for stfile in ('stagpy-paper.mplstyle',
                   'stagpy-slides.mplstyle'):
        stfile_conf = config.CONFIG_DIR / stfile
        if not (uptodate and stfile_conf.is_file()):
            stfile_local = pathlib.Path(__file__).parent / stfile
            shutil.copy(str(stfile_local), str(stfile_conf))


def load_mplstyle():
    """Try to load conf.plot.mplstyle matplotlib style."""
    plt = importlib.import_module('matplotlib.pyplot')
    if conf.plot.mplstyle:
        for style in conf.plot.mplstyle.split():
            found = False
            if not ISOLATED:
                stfile = config.CONFIG_DIR / (style + '.mplstyle')
                if stfile.is_file():
                    found = True
                    style = str(stfile)
            if not found:
                # try local version
                stfile = pathlib.Path(__file__).parent / (style + '.mplstyle')
                if stfile.is_file():
                    style = str(stfile)
            try:
                plt.style.use(style)
            except OSError:
                print(f'Cannot import style {style}.', file=sys.stderr)
                conf.plot.mplstyle = ''
    if conf.plot.xkcd:
        plt.xkcd()


if DEBUG:
    print('StagPy runs in DEBUG mode because the environment variable',
          'STAGPY_DEBUG is set to "True"', sep='\n', end='\n\n')
else:
    _PREV_INT = signal.signal(signal.SIGINT, sigint_handler)

try:
    __version__ = get_version(root='..', relative_to=__file__)
except LookupError:
    __version__ = get_distribution('stagpy').version
except (DistributionNotFound, ValueError):
    __version__ = 'unknown'

_CONF_FILES = ([config.CONFIG_FILE, config.CONFIG_LOCAL]
               if not ISOLATED else [])
conf = ConfigurationManager.from_dict_(config.CONF_DEF)
conf.set_config_files_(*_CONF_FILES)
if not ISOLATED:
    _check_config()
PARSING_OUT = conf.read_configs_()

load_mplstyle()

if not DEBUG:
    signal.signal(signal.SIGINT, _PREV_INT)
